import abc
from typing import Union
from pymongo import MongoClient
import pandas as pd
import json
import pytz
from datetime import datetime, timedelta, date
from ..utils import StatsDateTime, StatsProcessor, YoY_Calculator
import yaml


class StatsFetcher:

    def __init__(self, ticker, db_client):
        self.ticker = ticker
        self.db = db_client["company"]    # Replace with your database name
        self.collection = self.db["twse_stats"]

        self.timezone = pytz.timezone("Asia/Taipei")

        self.target_metric_dict = {
            'value': ['value'],
            'value_and_percentage': ['value', 'percentage'],
            'percentage': ['percentage'],
            'grand_total': ['grand_total'],
            'grand_total_values': ['grand_total', 'grand_total_percentage'],
            'grand_total_percentage': ['grand_total_percentage'],
            'growth': [f'YoY_{i}' for i in [1, 3, 5, 10]],
            'grand_total_growth': [f"YoY_{i}" for i in [1, 3, 5, 10]]
        }

    def prepare_query(self):
        return [
            {
                "$match": {
                    "ticker": self.ticker,
                }
            },
        ]

    def collect_data(self, start_date, end_date):
        pipeline = self.prepare_query()

        fetched_data = list(self.collection.aggregate(pipeline))

        return fetched_data[0]

    def str_to_datetime(self, date_str):
        year, month, day = [int(num) for num in date_str.split("-")]

        date = datetime.strptime(date_str, "%Y-%m-%d")
        date = self.timezone.localize(date)

        season = (month - 1) // 3 + 1

        return StatsDateTime(date, year, month, day, season)

    def has_required_columns(self, df: pd.DataFrame, required_cols=None):
        """
        Check if the required columns are present in the DataFrame.

        Args:
            df (pd.DataFrame): The DataFrame to check.
            required_cols (list, optional): List of required column names. 
                                            Defaults to ['date', 'open', 'high', 'low', 'close', 'volume'].

        Returns:
            bool: True if all required columns are present, False otherwise.
        """
        if required_cols is None:
            required_cols = ['date', 'open', 'high', 'low', 'close', 'volume']

        return all(col in df.columns for col in required_cols)


class BaseTEJFetcher(abc.ABC):

    @abc.abstractmethod
    def get(self):
        pass

    def get_latest_data_time(self, ticker):
        latest_data = self.collection.find_one(
            {
                "ticker": ticker
            }, 
            {
                "last_update": 1,
                "_id": 0
            }
        )

        try:
            latest_date = latest_data['last_update']["latest_data_date"]
        except Exception as e:
            latest_date = None

        return latest_date

    def process_value(self, value):
        if isinstance(value, str) and "%" in value:
            value = value.replace("%", "")
        try:
            return float(value)
        except (ValueError, TypeError):
            return None

    def calculate_growth(self, this_value, last_value, delta):
        try:
            return YoY_Calculator.cal_growth(this_value, last_value, delta) * 100
        except Exception:
            return None

    def cal_YoY(self, data_dict: dict, start_year: int, end_year: int, season: int):
        year_shifts = [1, 3, 5, 10]
        return_dict = {}
        
        for year in range(start_year, end_year + 1):
            year_data = data_dict.get(f"{year}Q{season}", {}).copy()
            if not year_data:
                continue
            
            for key in list(year_data.keys()):
                if key == "season":
                    continue
                
                this_value = self.process_value(year_data[key])
                if this_value is None:
                    year_data.pop(key)
                    continue
                
                temp_dict = {"value": year_data[key]}
                for shift in year_shifts:
                    past_year = year - shift
                    last_value = data_dict.get(f"{past_year}Q{season}", {}).get(key)
                    last_value = self.process_value(last_value)
                    growth = self.calculate_growth(this_value, last_value, shift) if last_value is not None else None

                    temp_dict[f"YoY_{shift}"] = (f"{growth:.2f}%" if growth else None)
                year_data[key] = temp_dict
            
            return_dict[f"{year}Q{season}"] = year_data
        
        return return_dict

    def cal_QoQ(self, data_dict):
        return_dict = {}
        
        for time_index, this_data in data_dict.items():
            year, season = map(int, time_index.split("Q"))
            last_year, last_season = (year - 1, 4) if season == 1 else (year, season - 1)
            
            for key in list(this_data.keys()):
                if key == "season":
                    continue
                
                this_value = self.process_value(this_data[key])
                if this_value is None:
                    this_data.pop(key)
                    continue
                
                temp_dict = {"value": this_data[key]}
                last_value = data_dict.get(f"{last_year}Q{last_season}", {}).get(key, {}).get('value')
                last_value = self.process_value(last_value)
                growth = self.calculate_growth(this_value, last_value, 1) if last_value is not None else None
                temp_dict['growth'] = (f"{growth:.2f}%" if growth else None)
                
                this_data[key] = temp_dict
            
            return_dict[time_index] = this_data
        
        return return_dict

    def get_dict_of_df(self, data_dict):
        """
        dict[dict] -> dict[df]
        """
        for key in data_dict.keys():
            data_dict[key] = pd.DataFrame.from_dict(data_dict[key])
        return data_dict

    def set_time_shift(self, date: Union[str, datetime], period: str):
        if isinstance(date, str):
            date = datetime.strptime(date, "%Y-%m-%d")

        period_mapping = {
            "1d": timedelta(days=1),
            "7d": timedelta(days=7),
            "1m": timedelta(days=30),
            "3m": timedelta(days=90),
            "1y": timedelta(days=365),
            "3y": timedelta(days=365 * 3),
            "5y": timedelta(days=365 * 5),
            "10y": timedelta(days=365 * 10),
        }

        if period == "all":
            return datetime.strptime("1991-01-01", "%Y-%m-%d")

        return date - period_mapping.get(period, timedelta(days=0))  # 預設為不變"



import pytest

mongo_uri = "mongodb+srv://admin:Neurowatt456&@axonnews-mongodb.nlui1.mongodb.net/?retryWrites=true&w=majority&appName=axonnews-mongodb"
ticker = "2330"
# collection_name = "TWN/AFESTM1"
collection_name = "TWN/AINVFQ1"

def test_QoQ():

    from neurostats_API import FinanceReportFetcher

    fetcher = FinanceReportFetcher(mongo_uri, collection_name=collection_name)

    modes = [
        fetcher.FetchMode.QOQ,
        fetcher.FetchMode.QOQ_NOCAL
    ]


    for mode in modes:
        data = fetcher.get(
            ticker,
            fetch_mode = mode,
            start_date="2024-01-01",
            end_date="2024-12-31",
            indexes = ["eps"]
        )
        print(data)
        if (isinstance(data, dict)):
            print(len(data.keys()))
        else:
            print(len(data.index))
def test_YoY():
    from neurostats_API import FinanceReportFetcher
    fetcher = FinanceReportFetcher(mongo_uri, collection_name=collection_name)

    modes = [
        fetcher.FetchMode.YOY,
        fetcher.FetchMode.YOY_NOCAL
    ]

    for mode in modes:
        data = fetcher.get(ticker, mode, indexes = ["r405"])
        print (data)

def test_stock_price():
    from neurostats_API import TEJStockPriceFetcher

    fetcher = TEJStockPriceFetcher(mongo_uri, collection_name="TWN/APIPRCD")

    data = fetcher.get(ticker, period='3y')

    print(data)



from importlib.resources import files
import json
import numpy as np
import pandas as pd
import yaml

target_metric_dict = {
    'value': ['value'],
    'value_and_percentage': ['value', 'percentage'],
    'percentage': ['percentage'],
    'grand_total': ['grand_total'],
    'grand_total_values': ['grand_total', 'grand_total_percentage'],
    'grand_total_percentage': ['grand_total_percentage'],
    'growth': [f'YoY_{i}' for i in [1, 3, 5, 10]],
    'grand_total_growth': [f"grand_total_YoY_{i}" for i in [1, 3, 5, 10]]
}


class StatsProcessor:
    """
    1. 讀檔: txt / yaml
    2. 將巢狀dictionary / DataFrame扁平化
    """

    @classmethod
    def load_txt(cls, filename, json_load=True):
        txt_path = files('neurostats_API.tools').joinpath(filename)
        with open(txt_path, 'r', encoding='utf-8') as f:
            data = json.load(f) if (json_load) else f.read()
        return data

    @classmethod
    def load_yaml(cls, filename):
        yaml_path = files('neurostats_API.tools').joinpath(filename)
        with open(yaml_path, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)

        return data

    @classmethod
    def expand_value_percentage(cls, dataframe):

        expanded_columns = {}
        for col in dataframe.columns:
            # Use json_normalize to split 'value' and 'percentage'
            expanded_df = pd.json_normalize(
                dataframe[col]).add_prefix(f"{col}_")
            expanded_df.index = dataframe.index
            # Append the expanded columns to the new DataFrame
            expanded_columns[col] = expanded_df

        expanded_df = pd.concat(expanded_columns.values(), axis=1)

        return expanded_df

    @classmethod
    def slice_table(
            cls,
            total_table,
            mode='value',
            target_index=None,  # None or Str， 要特別抓哪個index
    ):
        """
        total_table: column應為 <時間>_<單位>
        對只有單層column的table，切出想要的index
        """
        times = [
            column.split("_")[0] for column in total_table.columns.unique()
        ]  #取出timeIndex
        try:
            target_metrics = target_metric_dict[mode]
        except KeyError as e:
            return f"mode Error: Get mode should be {list(target_metric_dict.keys())} but get {mode}"

        desired_order = [
            f"{time}_{value_name}" for time in times
            for value_name in target_metrics
        ]

        if (target_index):
            target_index = target_index.split()
            sliced_table = total_table.loc[target_index, desired_order].T

            return sliced_table.T

        else:
            return total_table.loc[:, desired_order]

    @classmethod
    def slice_multi_col_table(
            cls,
            total_table,
            mode='value',
            target_index=None,  # None or Str， 要特別抓哪個index
    ):
        """
        對Multicolumn的dataframe切出目標的index
        """
        try:
            target_metrics = target_metric_dict[mode]
        except KeyError as e:
            return f"mode Error: Get mode should be {list(target_metric_dict.keys())} but get {mode}"

        times = total_table.columns.get_level_values(0).unique()
        desired_order = [(time, value_name) for time in times
                         for value_name in target_metrics]

        if (target_index):
            target_index = target_index.split()
            try:
                sliced_table = total_table.loc[
                    target_index,
                    pd.IndexSlice[:, target_metrics]][desired_order].T

            except Exception as e:  # 沒辦法完整取得表格
                # 先設立空表格
                empty_index = pd.Index(desired_order)
                empty_columns = pd.Index(target_index)
                sliced_table = pd.DataFrame(index=empty_index,
                                            columns=empty_columns)

                try:
                    # 提取有效的部分資料
                    partial_table = total_table.loc[
                        total_table.index.intersection(target_index),
                        pd.IndexSlice[:, target_metrics]
                    ]

                    # 遍歷 partial_table 的索引和值，手動填入 sliced_table
                    for row_index in partial_table.index:
                        for col_index in partial_table.columns:
                            if col_index in desired_order and row_index in target_index:
                                sliced_table.loc[col_index, row_index] = partial_table.loc[row_index, col_index]

                    # 確保 `sliced_table` 的排序符合 `desired_order`
                    sliced_table = sliced_table.reindex(index=desired_order, columns=target_index)
                except Exception as sub_e:
                    pass

            if (mode == 'value_and_percentage'):  # 因應balance_sheet 頁面的格式
                return_table = sliced_table.T
                return_table.columns = [
                    "_".join(flatten_indexs)
                    for flatten_indexs in return_table.columns.to_flat_index()
                ]
                return return_table

            sliced_table = sliced_table.reset_index()
            sliced_table = sliced_table.pivot(index='level_1',
                                              columns='level_0',
                                              values=target_index).sort_index(
                                                  axis=1,
                                                  level=1,
                                                  ascending=False)

            sliced_table.columns = sliced_table.columns.get_level_values(1)
            sliced_table.columns.name = None
            sliced_table.index.name = None

            return sliced_table.reindex(target_metrics)

        else:
            return_table = total_table.loc[:, pd.IndexSlice[:,
                                                            target_metrics]][
                                                                desired_order]
            return_table.columns = [
                "_".join(flatten_indexs)
                for flatten_indexs in return_table.columns.to_flat_index()
            ]
            return return_table

    @classmethod
    def cal_percentage(cls, value, postfix="%"):
        if (isinstance(value, (float, int))):
            value = np.round(value * 100, 2).item()
            value = f"{value:.2f}{postfix}"

            return value

        else:
            return value

    @classmethod
    def cal_non_percentage(cls, value, to_str=False, postfix="元"):
        if (isinstance(value, (float, int))):

            value = np.round(value, 2).item()
            if (postfix == "千元"):
                value *= 1000
                try:
                    value = int(value)
                except Exception as e:
                    pass

                postfix = "元"

            if (to_str):
                if (isinstance(value, float)):
                    try:
                        value = f"{value:.2f}{postfix}"
                    except Exception as e:
                        value = f"{value}{postfix}"
                    return value
                elif (isinstance(value, int)):
                    value = f"{value}{postfix}"
                    return value

            else:
                return value

        else:
            return value

    @classmethod
    def cal_round_int(cls, value):
        """
        將數值取到個位數後轉為int
        """
        if (isinstance(value, (int, float))):
            return int(np.round(value).item())
        else:
            return value
    
    @classmethod
    def list_of_dict_to_dict(
        cls,
        data_list: list,
        key: str = "",
        keys: list = [],
        delimeter: str = "_",
        data_key: str = "Q"
    ):
        """
        TEJ DB 用
        List[Dict] -> Dict[Dict]
        input:
        data_list(List):
            [
                { "data":
                    {
                        "year": 2021...
                        "season": 1,
                        "Q": {}...
                        
                    }
                }
            ]
        
        key(str): 選擇哪一個key作為轉化後的index
        delimeter(str): 多個key時要用甚麼分隔
        return: 
        {
            "2021" : {# Q下的資料} ...
        }

        or  (keys = ['year', 'season'])
        {
            "2021Q2" : {}
        }
        """
        assert (key or keys), "func list_of_dict_to_dict must have argument \"key\" or \"keys\""

        return_dict = {}
        if (key):
            keys = [key]
        for data in data_list:
            data = data['data']

            pop_keys = []

            for key in keys:
                assert (key in data.keys())
                pop_keys.append(str(data.pop(key)))
            
            pop_key = delimeter.join(pop_keys)
            return_dict[pop_key] = data[data_key]
        
        return return_dict


from .base import StatsFetcher, StatsDateTime
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import importlib.resources as pkg_resources
import yaml


class FinanceOverviewFetcher(StatsFetcher):
    """
    對應iFa.ai -> 財務分析 -> 重要指標(finance_overview)
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.target_fields = StatsProcessor.load_yaml(
            "finance_overview_dict.yaml")
        self.inverse_dict = StatsProcessor.load_txt(
            "seasonal_data_field_dict.txt", json_load=True)

    def prepare_query(self, target_year, target_season):

        pipeline = super().prepare_query()

        target_query = {
            "year": "$$target_season_data.year",
            "season": "$$target_season_data.season",
        }

        for key, target_sets in self.target_fields.items():
            try:
                small_targets = target_sets['field']

                value_index = target_sets['value']  # "金額" or "%"

                for small_target in small_targets:
                    big_target = self.inverse_dict[
                    small_target]  # balance_sheet/profit_lose/cash_flow
                    if (small_target == "利息費用_bank"):
                        small_target = small_target[:small_target.find("_bank")]
                    target_query.update({
                        f"{key}":
                        f"$$target_season_data.{big_target}.{small_target}.{value_index}"
                    })
            except Exception:
                continue

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "seasonal_data": {
                    "$map": {
                        "input": {
                            "$filter": {
                                "input": "$seasonal_data",
                                "as": "season",
                                "cond": {
                                    "$and": [{
                                        "$eq": ["$$season.year", target_year]
                                    }, {
                                        "$eq":
                                        ["$$season.season", target_season]
                                    }]
                                }
                            }
                        },
                        "as": "target_season_data",
                        "in": target_query
                    }
                }
            }
        })

        return pipeline

    def collect_data(self, target_year, target_season):
        pipeline = self.prepare_query(target_year, target_season)

        fetched_data = self.collection.aggregate(pipeline)

        fetched_data = list(fetched_data)

        return fetched_data[0]

    def query_data(self):

        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection)['last_update_time']
            year = latest_time['seasonal_data']['latest_year']
            season = latest_time['seasonal_data']['latest_season']
        except Exception as e:
            today = StatsDateTime.get_today()
            year = today.year - 1 if (today.season == 1) else today.year
            season = 4 if (today.season == 1) else today.season - 1

        fetched_data = self.collect_data(year, season)
        finance_dict = fetched_data['seasonal_data'][0]
        FinanceOverviewProcessor.process_rate(finance_dict)
        FinanceOverviewProcessor.process_all(finance_dict)
        self.fill_nan_index(finance_dict)
        FinanceOverviewProcessor.process_thousand_dollar(finance_dict)
        fetched_data['seasonal_data'] = finance_dict

        return fetched_data
    
    def fill_nan_index(self, finance_dict):
        for key in self.target_fields.keys():
            if (key not in finance_dict.keys()):
                finance_dict[key] = None


class FinanceOverviewProcessor(StatsProcessor):

    @classmethod
    def process_rate(cls, finance_dict):
        for key in finance_dict:
            if ('YoY' in key):
                finance_dict[key] = StatsProcessor.cal_percentage(
                    finance_dict[key])
            elif ("rate" in key or 'ratio' in key):
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key], to_str=True, postfix='%')
            else:
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key])
    
    
    @classmethod
    def process_thousand_dollar(cls, finance_dict):
        process_index = [
            "revenue",
            "gross_profit",
            "operating_income",
            "net_income",
            "operating_cash_flow",
            "invest_cash_flow",
            "financing_cash_flow",
            "fcf",

            'current_assets',
            'current_liabilities',
            'non_current_assets',
            'non_current_liabilities',
            'total_assets',
            "total_liabilities",
            "equity"
        ]
        
        for index in process_index:
            try:
                finance_dict[index] = StatsProcessor.cal_non_percentage(finance_dict[index], postfix="千元")
            except Exception as e:
                finance_dict[index] = None
            

    @classmethod
    def process_all(cls, finance_dict):
        methods = [
            cls.cal_EBIT, cls.cal_share_outstanding, cls.cal_fcf,
            cls.cal_interest_bearing_debt, cls.cal_revenue_per_share,
            cls.cal_gross_per_share, cls.cal_operating_income_per_share,
            cls.cal_operating_cash_flow_per_share, cls.fcf_per_share,
            cls.cal_roa, cls.cal_roe, cls.cal_gross_over_asset, cls.cal_roce,
            cls.cal_gross_profit_marginal, cls.cal_operation_profit_rate,
            cls.cal_operating_cash_flow_profit_rate, cls.cal_dso,
            cls.cal_account_receive_over_revenue, cls.cal_dpo,
            cls.cal_inventories_cycle_ratio, cls.cal_dio,
            cls.cal_inventories_revenue_ratio,
            cls.cal_cash_of_conversion_cycle, cls.cal_asset_turnover,
            cls.cal_application_turnover, cls.cal_current_ratio,
            cls.cal_quick_ratio, cls.cal_debt_to_equity_ratio,
            cls.cal_net_debt_to_equity_ratio, cls.cal_interest_coverage_ratio,
            cls.cal_debt_to_operating_cash_flow,
            cls.cal_debt_to_free_cash_flow, cls.cal_cash_flow_ratio,
        ]

        for method in methods:
            method(finance_dict)


    @classmethod
    def cal_EBIT(cls, finance_dict):
        """
        計算EBIT
        EBIT = 營業收入 - 營業成本 - 營業費用 - 所得稅費用
        """
        try:
            EBIT = (finance_dict['revenue'] - finance_dict['operating_cost'] -
                    finance_dict['operating_expenses'] -
                    finance_dict['tax_fee'])
            finance_dict['EBIT'] = StatsProcessor.cal_non_percentage(EBIT)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['EBIT'] = None
            print(f"Error calculating EBIT: {e}")

    @classmethod
    def cal_fcf(cls, finance_dict):
        """
        計算自由現金流(FCF):
        自由現金流 =  營業現金流 + 投資現金流
        """
        try:
            fcf = (finance_dict["operating_cash_flow"] +
                   finance_dict["financing_cash_flow"])
            finance_dict["fcf"] = StatsProcessor.cal_non_percentage(fcf)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf'] = None
            print(f"Error calculating FCF: {e}")

    @classmethod
    def cal_interest_bearing_debt(cls, finance_dict):
        """
        計算有息負債
        短期借款+長期借款
        """
        finance_dict['interest_bearing_debt'] = 0.0

        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'short_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0
        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'long_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0

    @classmethod
    def cal_share_outstanding(cls, finance_dict):
        """
        計算流通股數
        流通股數 = 本期淨利 ÷ 基本每股盈餘
        """
        try:
            finance_dict["share_outstanding"] = (finance_dict['net_income'] /
                                                 finance_dict['eps'])
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['share_outstanding'] = None
            print(f"share_outstanding failed because of {str(e)}")

    @classmethod
    def cal_revenue_per_share(cls, finance_dict):
        """
        計算每股營收
        每股營收 = 營業收入 / 在外流通股數
        """
        try:
            revenue_per_share = (finance_dict['revenue'] /
                                 finance_dict['share_outstanding'])
            finance_dict[
                'revenue_per_share'] = StatsProcessor.cal_non_percentage(
                    revenue_per_share, False)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['revenue_per_share'] = None
            print(f"revenue_per_share failed because of {str(e)}")

    @classmethod
    def cal_gross_per_share(cls, finance_dict):
        """
        計算每股毛利
        = （當期營業毛利）÷（當期在外流通股數）
        """
        if ('gross_profit' not in finance_dict.keys()):
            try:
                finance_dict['gross_profit'] = (
                    finance_dict['revenue'] - 
                    finance_dict['operating_cost']
                )
            except:
                finance_dict['gross_profit'] = None
        try:
            gross_per_share = (finance_dict['gross_profit'] /
                               finance_dict['share_outstanding'])
            finance_dict[
                'gross_per_share'] = StatsProcessor.cal_non_percentage(
                    gross_per_share, False)

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_per_share'] = None
            print(f"gross_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_income_per_share(cls, finance_dict):
        """
        計算每股營業利益
        每股營業利益= （當期營業利益）÷（當期在外流通股數）
        """
        try:
            operating_income_per_share = (finance_dict['operating_income'] /
                                          finance_dict['share_outstanding'])
            finance_dict[
                'operating_income_per_share'] = StatsProcessor.cal_non_percentage(
                    operating_income_per_share)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_income_per_share'] = None
            print(f"operating_income_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_per_share(cls, finance_dict):
        """
        計算每股營業現金流
        = (當期營業現金流) ÷（當期在外流通股數）
        """
        try:
            operating_cash_flow_per_share = (
                finance_dict["operating_cash_flow"] /
                finance_dict['share_outstanding'])
            finance_dict[
                "operating_cash_flow_per_share"] = StatsProcessor.cal_non_percentage(
                    operating_cash_flow_per_share)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_cash_flow_per_share'] = None
            # print(f'operating_cash_flow_per_share because of {str(e)}')

    @classmethod
    def fcf_per_share(cls, finance_dict):
        """
        計算每股自由現金流
        每股自由現金流 = (當期自由現金流) ÷（當期在外流通股數）
        """
        try:
            fcf_per_share = (finance_dict['fcf'] /
                             finance_dict['share_outstanding'])
            finance_dict['fcf_per_share'] = StatsProcessor.cal_non_percentage(
                fcf_per_share)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf_per_share'] = None
            print(f"fcf_per_share failed because of {str(e)}")

# 盈利能力

    @classmethod
    def cal_roa(cls, finance_dict):
        """
        計算資產報酬率(ROA)
        ROA = [ 本期淨利 + 利息費用 × (1-有效稅率) ] ÷（資產總額）
        """
        try:
            roa = (
                finance_dict['net_income'] + finance_dict['interest'] +
                (1 * 0.1)  # 有效稅率需要改，這裡先設0.1
            ) / finance_dict['inventories']

            finance_dict["roa"] = StatsProcessor.cal_percentage(roa)
        except Exception as e:
            finance_dict["roa"] = None

    @classmethod
    def cal_roe(cls, finance_dict):
        """
        計算股東權益報酬率(ROE)
        ROE = (本期淨利) ÷（權益總額）
        """
        try:
            roe = (finance_dict['net_income'] / finance_dict['equity'])
            finance_dict['roe'] = StatsProcessor.cal_percentage(roe)
        except Exception as e:
            finance_dict['roe'] = None

    @classmethod
    def cal_gross_over_asset(cls, finance_dict):
        """
        計算營業毛利/總資產
        """
        try:
            gross_over_asset = (finance_dict['gross_profit'] /
                                finance_dict['total_assets'])
            finance_dict['gross_over_asset'] = StatsProcessor.cal_percentage(
                gross_over_asset)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_over_asset'] = None
            print(f"營業毛利/總資產 failed because of {str(e)}")

    @classmethod
    def cal_roce(cls, finance_dict):
        """
        計算資本運用報酬率(ROCE):
        ROCE = (稅前淨利＋利息費用) / (資產總額－流動負債)
        """
        try:
            roce = ((finance_dict['net_income_before_tax'] +
                     finance_dict['interest']) /
                    (finance_dict['total_assets'] -
                     finance_dict['current_liabilities']))
            finance_dict['roce'] = StatsProcessor.cal_percentage(roce)

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['roce'] = None
            print(f"ROCE failed because of {str(e)}")

    @classmethod
    def cal_gross_profit_marginal(cls, finance_dict):
        """
        計算營業毛利率(gross profit margin)
        營業毛利率 = 營業毛利 ÷ 營業收入
        """
        try:
            gross_profit_margin = (finance_dict['gross_profit'] /
                                   finance_dict['revenue'])
            finance_dict[
                'gross_profit_margin'] = StatsProcessor.cal_percentage(
                    gross_profit_margin)
        except Exception as e:
            finance_dict['gross_profit_margin'] = None
            print(f"gross_profit_margin failed because of {str(e)}")

    @classmethod
    def cal_operation_profit_rate(cls, finance_dict):
        """
        計算營業利益率
        營業利益率 = ( 營業收入－營業成本－營業費用）÷ 營業收入
        """
        try:
            operation_profit_rate = (
                finance_dict['revenue'] - finance_dict['operating_cost'] -
                finance_dict['operating_expenses']) / finance_dict['revenue']
            finance_dict[
                "operation_profit_rate"] = StatsProcessor.cal_percentage(
                    operation_profit_rate)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operation_profit_rate"] = None
            print(f"operation_profit failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_profit_rate(cls, finance_dict):
        """
        計算營業現金流利潤率
        營業現金流利潤率 = 營業活動現金流 ÷ 營業收入
        """
        try:
            operating_cash_flow_profit_rate = (
                finance_dict["operating_cash_flow"] / finance_dict["revenue"])
            finance_dict[
                "operating_cash_flow_profit_rate"] = StatsProcessor.cal_percentage(
                    operating_cash_flow_profit_rate)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operating_cash_flow_profit_rate"] = None
            print(
                f"operating_cash_flow_profit_rate failed because of {str(e)}")


# 成長動能

    """
    前四個已經有了 revenue_YoY, gross_prof_YoY, operating_income_YoY, net_income_YoY:
    後四個在金流，還需要處理
    """
    # 營運指標

    @classmethod
    def cal_dso(cls, finance_dict):
        """
        計算應收帳款收現天數(DSO)
        DSO = 365 × (應收帳款平均餘額 ÷ 營業收入)
        """
        try:
            dso = (365 *
                   (finance_dict['account_pay'] / finance_dict['revenue']))
            finance_dict['dso'] = StatsProcessor.cal_non_percentage(
                dso, to_str=True, postfix="日")
        except Exception as e:
            finance_dict['dso'] = None
            print(f"Error calculating 應收帳款收現天數 because of {str(e)}")

    @classmethod
    def cal_account_receive_over_revenue(cls, finance_dict):
        """
        計算應收帳款佔營收比率
        = 應收帳款平均餘額 ÷ 營業收入
        """
        try:
            account_receive_over_revenue = (finance_dict['account_receive'] /
                                            finance_dict['revenue'])
            finance_dict[
                "account_receive_over_revenue"] = StatsProcessor.cal_percentage(
                    account_receive_over_revenue)
        except Exception as e:
            finance_dict[
                "account_receive_over_revenue"] = None

    @classmethod
    def cal_dpo(cls, finance_dict):
        """
        計算應付帳款週轉天數
        DPO = 365天 ÷ (銷貨成本÷平均應付帳款)
        """
        try:
            dpo = (
                365 *
                (finance_dict['account_pay'] / finance_dict['operating_cost']))
            finance_dict["dpo"] = StatsProcessor.cal_non_percentage(
                dpo, to_str=True, postfix="日")
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dpo"] = None
            print(f"應付帳款週轉天數 failed because of {str(e)}")

    @classmethod
    def cal_inventories_cycle_ratio(cls, finance_dict):
        """
        計算存貨周轉率
        = 銷貨成本 ÷ 存貨
        """
        try:
            inventories_cycle_ratio = (finance_dict['operating_cost'] /
                                       finance_dict['inventories'])

            finance_dict[
                "inventories_cycle_ratio"] = StatsProcessor.cal_percentage(
                    inventories_cycle_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_cycle_ratio"] = None
            print(f"Error calculating 存貨周轉率 because of {str(e)}")

    @classmethod
    def cal_dio(cls, finance_dict):
        """
        計算 存貨週轉天數 or 平均售貨天數
        DIO = 365天 * (存貨 ÷ 銷貨成本)
        MUDA MUDA MUDA !!!
        """
        try:
            dio = 365 * (finance_dict["inventories"] /
                         finance_dict["operating_cost"])
            finance_dict["dio"] = StatsProcessor.cal_non_percentage(
                dio, to_str=True, postfix="日")
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dio"] = None
            print(f"Error calculating 存貨週轉天數 because of {str(e)}")

    @classmethod
    def cal_inventories_revenue_ratio(cls, finance_dict):
        """
        計算存貨佔營收比率
        存貨佔營收比= 存貨 ÷ 營業收入
        """
        try:
            inventories_revenue_ratio = (finance_dict['inventories'] /
                                         finance_dict['revenue'])

            finance_dict[
                "inventories_revenue_ratio"] = StatsProcessor.cal_percentage(
                    inventories_revenue_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_revenue_ratio"] = None
            print(f"Error calculating 存貨佔營收比率 because of {str(e)}")

    @classmethod
    def cal_cash_of_conversion_cycle(cls, finance_dict):
        """
        計算現金循環週期
        存貨週轉天數 + 應收帳款週轉天數 - 應付帳款週轉天數
        """
        try:
            cash_of_conversion_cycle = (finance_dict["dio"] +
                                        finance_dict["dso"] -
                                        finance_dict['dpo'])
            finance_dict[
                "cash_of_conversion_cycle"] = StatsProcessor.cal_non_percentage(
                    cash_of_conversion_cycle, to_str=True, postfix="日")
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["cash_of_conversion_cycle"] = None

    @classmethod
    def cal_asset_turnover(cls, finance_dict):
        """
        計算資產周轉率
        營業收入 ÷ 資產總額
        """
        try:
            asset_turnover = (finance_dict["revenue"] /
                            finance_dict["inventories"])
            finance_dict["asset_turnover"] = StatsProcessor.cal_percentage(
                asset_turnover)
        except Exception as e:
            finance_dict["asset_turnover"] = None

    @classmethod
    def cal_application_turnover(cls, finance_dict):
        """
        不動產、廠房及設備週轉率
        營業收入 ÷ 不動產、廠房與設備平均餘額
        """
        try:
            applcation_turnover = (finance_dict['revenue'] /
                                   finance_dict["application"])
            finance_dict[
                'application_turnover'] = StatsProcessor.cal_percentage(
                    applcation_turnover)

        except Exception as e:
            finance_dict['application_turnover'] = None
        

    @classmethod
    def cal_current_ratio(cls, finance_dict):
        """
        計算流動比率 = 流動資產 / 流動負債
        """
        try:
            current_ratio = (finance_dict['current_assets'] /
                             finance_dict['current_liabilities'])
            finance_dict['current_ratio'] = StatsProcessor.cal_percentage(
                current_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['current_ratio'] = None
            print(f"Error calculating current ratio: {e}")

    @classmethod
    def cal_quick_ratio(cls, finance_dict):
        """
        速動比率
        (流動資產 - 存貨) / 流動負債
        """
        try:
            quick_ratio = (finance_dict['current_assets'] -
                           finance_dict['inventories']
                           ) / finance_dict['current_liabilities']
            finance_dict['quick_ratio'] = StatsProcessor.cal_percentage(
                quick_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['quick_ratio'] = None
            print(f"Error calculating quick ratio: {e}")

    @classmethod
    def cal_debt_to_equity_ratio(cls, finance_dict):
        """
        # 負債權益比率 = 總負債 / 股東權益
        """
        try:
            debt_to_equity_ratio = finance_dict[
                'total_liabilities'] / finance_dict['equity']
            finance_dict[
                'debt_to_equity_ratio'] = StatsProcessor.cal_percentage(
                    debt_to_equity_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_equity_ratio'] = None
            print(f"Error calculating debt to equity ratio: {e}")

    @classmethod
    def cal_net_debt_to_equity_ratio(cls, finance_dict):
        """
        # 淨負債權益比率 = (總負債 - 現金及約當現金) / 股東權益
        """
        try:
            net_debt_to_equity_ratio = (
                finance_dict['total_liabilities'] -
                finance_dict['cash_and_cash_equivalents']
            ) / finance_dict['equity']
            finance_dict[
                'net_debt_to_equity_ratio'] = StatsProcessor.cal_percentage(
                    net_debt_to_equity_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['net_debt_to_equity_ratio'] = None
            print(f"Error calculating net debt to equity ratio: {e}")

    @classmethod
    def cal_interest_coverage_ratio(cls, finance_dict):
        """
        # 利息保障倍數 = EBIT / 利息費用
        """
        try:
            interest_coverage_ratio = finance_dict['EBIT'] / finance_dict[
                'interest_expense']
            finance_dict[
                'interest_coverage_ratio'] = StatsProcessor.cal_non_percentage(
                    interest_coverage_ratio, to_str=True, postfix="倍")
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_coverage_ratio'] = None
            print(f"Error calculating interest coverage ratio: {e}")

    @classmethod
    def cal_debt_to_operating_cash_flow(cls, finance_dict):
        """
        有息負債 / 營業活動現金流
        """
        try:
            debt_to_operating_cash_flow = finance_dict[
                'interest_bearing_debt'] / finance_dict['operating_cash_flow']
            finance_dict[
                'debt_to_operating_cash_flow'] = StatsProcessor.cal_percentage(
                    debt_to_operating_cash_flow)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_operating_cash_flow'] = None
            print(f"Error calculating debt to operating cash flow: {e}")

    @classmethod
    def cal_debt_to_free_cash_flow(cls, finance_dict):
        """
        # 有息負債 / 自由現金流
        """
        try:
            debt_to_free_cash_flow = finance_dict[
                'interest_bearing_debt'] / finance_dict['fcf']
            finance_dict[
                'debt_to_free_cash_flow'] = StatsProcessor.cal_percentage(
                    debt_to_free_cash_flow)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_free_cash_flow'] = None
            print(f"Error calculating debt to free cash flow: {e}")

    @classmethod
    def cal_cash_flow_ratio(cls, finance_dict):
        """
        # 現金流量比率 = 營業活動現金流 / 流動負債
        """
        try:
            cash_flow_ratio = finance_dict[
                'operating_cash_flow'] / finance_dict['current_liabilities']
            finance_dict['cash_flow_ratio'] = StatsProcessor.cal_percentage(
                cash_flow_ratio)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['cash_flow_ratio'] = None
            print(f"Error calculating cash flow ratio: {e}")



from datetime import datetime

class StatsDateTime():

    def __init__(self, date, year, month, day, season):
        self.date = date
        self.year = year
        self.month = month
        self.day = day
        self.season = season
    
    @classmethod
    def get_today(cls):
        today = datetime.today()
        this_year = today.year
        this_month = today.month
        this_day = today.day
        this_season = ((this_month - 1) // 3) + 1

        return StatsDateTime(today, this_year, this_month, this_day,
                             this_season)

    @classmethod
    def get_latest_time(cls, ticker, collection):
        return collection.find_one(
            { "ticker" : ticker }, 
            { "_id": 0, "last_update_time": 1 }
        )




import argparse
import pprint
from fetchers.value_invest import ValueFetcher
import pandas as pd
pp = pprint.PrettyPrinter(
    indent = 2
)

pd.options.display.max_rows = 4
pd.options.display.max_columns = 4

def main():
    parser = argparse.ArgumentParser(
        prog="Neurostats Datasets fetching",
        description="Main function of fetching Neurostat Dataset",
    )    
    
    parser.add_argument("--ticker", type=str, default="2330")
    parser.add_argument("--start_date", type=str, default="2018-01-01")
    parser.add_argument("--end_date", type=str, default="2199-12-31")

    args = parser.parse_args()

    fetcher = ValueFetcher(args.ticker)

    query_data = fetcher.query_data()
    print("{")
    for key, value in query_data.items():
        print(f"\t\"{key}\":")

        print(f"{value}")
    print("}")

if (__name__ == "__main__"):
    main()


import pandas as pd
import json
import pytz
from datetime import datetime, timedelta, date
from .data_process import StatsProcessor
import yaml


class StatsFetcher:

    def __init__(self, db_client):
        self.db = db_client["company"]  # Replace with your database name
        self.collection = self.db["twse_stats"]

        self.timezone = pytz.timezone("Asia/Taipei")

        self.inverse_dict = StatsProcessor.load_txt("seasonal_data_field_dict.txt", json_load=True)

        self.seasons = ["01", "02", "03", "04"]

        self.pipeline = list()

        self.target_metric_dict = {
            'value': ['value'],
            'value_and_percentage': ['value', 'percentage'],
            'percentage': ['percentage'],
            'grand_total': ['grand_total'],
            'grand_total_values':['grand_total', 'grand_total_percentage'],
            'grand_total_percentage':['grand_total_percentage'],
            'growth': [f'YoY_{i}' for i in [1,3,5,10]],
            'grand_total_growth': [f"YoY_{i}" for i in [1,3,5,10]]
        }

        self.__return_dict = dict()

    def _flush_dict(self):
        self.__return_dict = dict()

    def _default_query(self, ticker, start_date, end_date):

        start_year, start_month, start_day = [
            int(num) for num in start_date.split("-")
        ]
        end_year, end_month, end_day = [
            int(num) for num in end_date.split("-")
        ]

        start_date = datetime.strptime(start_date, "%Y-%m-%d")
        end_date = datetime.strptime(end_date, "%Y-%m-%d")
        start_date = self.timezone.localize(start_date)
        end_date = self.timezone.localize(end_date)

        start_season = start_month // 3 + 1
        end_season = end_month // 3 + 1

        ticker = ticker.strip().split()[0]

        self.pipeline = [
            # 1. Find Ticker
            {
                "$match": {
                    "ticker": ticker,
                }
            },
            # 2. Find by date
            {
                "$project": {
                    "ticker": 1,
                    "company_name": 1,
                    # 2.1 Filter monthly_data
                    "daily_data": {
                        "$filter": {
                            "input": "$daily_data",
                            "as": "daily",
                            "cond": {
                                "$and": [
                                    {
                                        "$gte": ["$$daily.date", start_date]
                                    },
                                    {
                                        "$lte": ["$$daily.date", end_date]
                                    },
                                ]
                            },
                        }
                    },
                    # 2.2 Filter monthly_data
                    "monthly_data": {
                        "$filter": {
                            "input": "$monthly_data",
                            "as": "monthly",
                            "cond": {
                                "$or": [
                                    {
                                        "$and": [
                                            {
                                                "$eq":
                                                ["$$monthly.year", start_year]
                                            },
                                            {
                                                "$gte": [
                                                    "$$monthly.month",
                                                    start_month
                                                ]
                                            },
                                        ]
                                    },
                                    {
                                        "$and": [
                                            {
                                                "$eq":
                                                ["$$monthly.year", end_year]
                                            },
                                            {
                                                "$lte":
                                                ["$$monthly.month", end_month]
                                            },
                                        ]
                                    },
                                    {
                                        "$and": [
                                            {
                                                "$gt":
                                                ["$$monthly.year", start_year]
                                            },
                                            {
                                                "$lt":
                                                ["$$monthly.year", end_year]
                                            },
                                        ]
                                    },
                                ]
                            },
                        }
                    },
                    # 2.3 Filter seasonal_data
                    "seasonal_data": {
                        "$filter": {
                            "input": "$seasonal_data",
                            "as": "seasonal",
                            "cond": {
                                "$or": [
                                    {
                                        "$and": [
                                            {
                                                "$eq": [
                                                    "$$seasonal.year",
                                                    start_year
                                                ]
                                            },
                                            {
                                                "$gte": [
                                                    "$$seasonal.season",
                                                    start_season
                                                ]
                                            },
                                        ]
                                    },
                                    {
                                        "$and": [
                                            {
                                                "$eq":
                                                ["$$seasonal.year", end_year]
                                            },
                                            {
                                                "$lte": [
                                                    "$$seasonal.season",
                                                    end_season
                                                ]
                                            },
                                        ]
                                    },
                                    {
                                        "$and": [
                                            {
                                                "$gt": [
                                                    "$$seasonal.year",
                                                    start_year
                                                ]
                                            },
                                            {
                                                "$lt":
                                                ["$$seasonal.year", end_year]
                                            },
                                        ]
                                    },
                                ]
                            },
                        }
                    },
                    "yearly_data": {
                        "$filter": {
                            "input": "$yearly_data",
                            "as": "yearly",
                            "cond": {
                                "$and": [
                                    {
                                        "$gte": ["$$yearly.year", 107]
                                    },
                                    {
                                        "$lte": ["$$yearly.year", end_year]
                                    },
                                ]
                            },
                        }
                    }
                }
            },
        ]

    def query_data(self, ticker, start_date, end_date):
        """
        Return : Dict {
                    'ticker' : <ticker>,
                    'company_name': <company_name>,
                    'daily_data': List[Dict]
                    'monthly_data': List[Dict]
                    'seasonal_data': List[Dict]
                }
        """

        self._default_query(ticker, start_date, end_date)

        fetched_datas = list(self.collection.aggregate(self.pipeline))

        return fetched_datas[0]

    def query_values(self, ticker, start_date, end_date):
        self._default_query(ticker, start_date, end_date)

        self.pipeline.append({
            "$project": {
                "ticker": 1,
                "company_name": 1,

                # Transform daily_data to include only index and date
                "daily_data": {
                    "$map": {
                        "input": "$daily_data",
                        "as": "daily_item",
                        "in": {
                            "date": "$$daily_item.date",
                            "close": "$$daily_item.close",
                            "P_B": "$$daily_item.P_B",
                            "P_E": "$$daily_item.P_E",
                            "P_FCF": "$$daily_item.P_FCF",
                            "P_S": "$$daily_item.P_S",
                            "EV_OPI": "$$daily_item.EV_OPI",
                            "EV_EBIT": "$$daily_item.EV_EBIT",
                            "EV_EBITDA": "$$daily_item.EV_EBITDA",
                            "EV_S": "$$daily_item.EV_S"
                        }
                    }
                },
                "yearly_data": 1
            }
        })

        fetched_datas = list(self.collection.aggregate(self.pipeline))

        return fetched_datas[0]

    def query_stock_price(self, ticker, start_date, end_date):

        self.pipeline.append({
            "$project": {
                "ticker": 1,
                "company_name": 1,

                # Transform daily_data to include only index and date
                "daily_data": {
                    "$map": {
                        "input": "$daily_data",
                        "as": "daily_item",
                        "in": {
                            "date": "$$daily_item.date",
                            "open": "$$daily_item.open",
                            "high": "$$daily_item.high",
                            "los": "$$daily_item.low",
                            "close": "$$daily_item.close",
                        }
                    }
                },
            }
        })

        fetched_datas = list(self.collection.aggregate(self.pipeline))

        return fetched_datas[0]

    def query_seasonal_data(self, ticker, start_date, end_date, sheet, target_season):

        self._default_query(ticker, start_date, end_date)
        self.pipeline.append({
            "$project": {
                "ticker": 1,
                "company_name": 1,
                "seasonal_data": {
                    "$filter":{
                        "input":{
                            "$map": {
                                "input": "$seasonal_data",
                                "as": "seasonal_item",
                                "in": {
                                    "year": "$$seasonal_item.year",
                                    "season": "$$seasonal_item.season",
                                    f"{sheet}": f"$$seasonal_item.{sheet}"
                                }
                            }
                        },
                        "as": "filtered_item",
                        "cond": { "$eq": ["$$filtered_item.season", target_season] }
                    }
                },
            }
        })

        self.pipeline.append({"$unwind": "$seasonal_data"})

        self.pipeline.append(
            {"$sort": {
                "seasonal_data.year": 1,
                "seasonal_data.season": 1
            }})

        fetched_datas = list(self.collection.aggregate(self.pipeline))

        return fetched_datas

    def query_month_revenue(self, ticker, start_date, end_date):
        self._default_query(ticker, start_date, end_date)
        self.pipeline.append(
            {
                "$project": {
                    "ticker": 1,
                    "company_name": 1,
                    "monthly_data": {
                        "$sortArray": {
                            "input": "$monthly_data",
                            "sortBy": {
                                "year": 1,
                                "month": 1
                            }
                        }
                    },
                }
            }, )

        fetched_datas = list(self.collection.aggregate(self.pipeline))

        return fetched_datas[0]

    def query_latest_values(self, ticker):
        """
        傳回最近一天的價值面

        return : Dict {
            "ticker": 股票代碼,
            "company_name": 公司中文名稱,
        ## 以下八個是iFa項目
            "P_E": 本益比,
            "P_B": 股價,
            "P_FCF": 股價自由現金流比,
            "P_S": 股價營收比,
            "EV_EBIT: ,
            "EV_EBITDA": ,
            "EV_OPI": ,
            "EV_S"; ,
        ## 以上八個是iFa項目
            "close": 收盤價,
            "EV": 市場價值
        }
        """
        today = date.today()
        yesterday = (today - timedelta(days=14)).strftime("%Y-%m-%d")
        today = today.strftime("%Y-%m-%d")

        fetched_datas = self.query_values(ticker, yesterday, today)

        daily_data = fetched_datas.pop('daily_data')
        fetched_datas.update(daily_data[-1])
        return fetched_datas

    def query_latest_month_revenue(self, ticker):
        """
        傳回最新一期的月營收
        """

        today = date.today()

        last_month = (today - timedelta(days=30)).strftime("%Y-%m-%d")
        today = today.strftime("%Y-%m-%d")

        fetched_datas = self.query_month_revenue(ticker, last_month, today)

        print(fetched_datas)

        latest_month_revenue = fetched_datas['monthly_data']
        fetched_datas.pop('monthly_data')

        fetched_datas.update(latest_month_revenue[-1])

        return fetched_datas

    def query_latest_seasonal_data(self, ticker):
        """
        傳回最新一期的季報
        """
        today = date.today()

        last_season = (today - timedelta(days=90)).strftime("%Y-%m-%d")
        today = today.strftime("%Y-%m-%d")

        fetched_datas = self.query_seasonal_data(ticker, last_season, today)

        print(fetched_datas)

        latest_seasonal_data = fetched_datas['seasonal_data']
        fetched_datas.pop('seasonal_data')

        fetched_datas.update(latest_seasonal_data[-1])

        return fetched_datas

    def get_value_sheet(self, ticker):
        """
        iFa.ai: 價值投資-> 市場指標
        """
        """
        傳回最近一天的價值面

        return : Dict {
            "ticker": 股票代碼,
            "company_name": 公司中文名稱,
            "daily_data":{
            ## 以下八個是iFa項目
                "P_E": 本益比,
                "P_B": 股價,
                "P_FCF": 股價自由現金流比,
                "P_S": 股價營收比,
                "EV_EBIT: ,
                "EV_EBITDA": ,
                "EV_OPI": ,
                "EV_S"; 
            ## 以上八個是iFa項目
                "close": 收盤價,
            }

            "yearly_data": pd.DataFrame (下表格為範例)
                year    P_E       P_FCF   P_B        P_S     EV_OPI    EV_EBIT   EV_EBITDA       EV_S
            0   107  16.68   29.155555  3.71  11.369868  29.837201  28.798274  187.647704  11.107886
            1   108  26.06   67.269095  5.41  17.025721  50.145736  47.853790  302.526388  17.088863
            2   109  27.98   95.650723  7.69  22.055379  53.346615  51.653834  205.847232  22.481951
            3   110  27.83  149.512474  7.68  22.047422  55.398018  54.221387  257.091893  22.615355
            4   111  13.11   48.562021  4.25  11.524975  24.683850  24.226554   66.953260  12.129333
            5   112  17.17  216.371410  4.59  16.419533  40.017707  37.699267  105.980652  17.127656
        }
        """
        today = date.today()
        this_year = today.year - 1911
        yesterday = (today - timedelta(days=14)).strftime("%Y-%m-%d")
        today = today.strftime("%Y-%m-%d")

        fetched_datas = self.query_values(ticker, yesterday, today)

        fetched_datas['daily_data'] = fetched_datas['daily_data'][-1]

        latest_data = {"year": f"過去4季"}

        latest_data.update(fetched_datas['daily_data'])
        latest_data.pop("date")
        latest_data.pop("close")

        fetched_datas['yearly_data'].append(latest_data)

        fetched_datas['yearly_data'] = pd.DataFrame.from_dict(
            fetched_datas['yearly_data'])

        return fetched_datas

    def get_month_revenue_sheet(self, ticker):
        """
        iFa.ai: 財務分析 -> 每月營收

        return: Dict {
            'ticker': str,
            'company_name': str,
            'month_revenue': pd.DataFrame  (歷年的月營收以及到今年最新月份累計的月營收表格)
            'this_month_revenue_over_years': pd.DataFrame (今年這個月的月營收與歷年同月份的營收比較)
            'grand_total_over_years': pd.DataFrame (累計至今年這個月的月營收與歷年的比較)
        }
        """

        today = datetime.today()
        today = today.strftime("%Y-%m-%d")

        start_date = "2014-01-01"

        query_data = self.query_month_revenue(ticker, start_date, today)

        monthly_data = query_data['monthly_data']

        this_month = monthly_data[-1]["month"]

        month_dict = {i: None for i in range(1, 13)}
        month_dict[f"grand_total"] = None

        monthly_dict = dict()

        revenue_by_year = dict()
        single_month_revenue_dict = {
            "revenue": None,
            "MoM": None,
            "YoY": None,
            "YoY_3": None,
            "YoY_5": None,
            "YoY_10": None
        }

        grand_total_by_year = dict()
        grand_total_dict = {
            "revenue": None,
            "MoM": None,
            "YoY": None,
            "YoY_3": None,
            "YoY_5": None,
            "YoY_10": None
        }

        for data in monthly_data:
            try:
                monthly_dict[data['year']][data['month']] = data['revenue']
            except:
                monthly_dict[data['year']] = month_dict.copy()
                monthly_dict[data['year']][data['month']] = data['revenue']

            try:
                if (data['last_year_revenue']
                        != monthly_dict[data['year'] - 1][data['month']]):
                    monthly_dict[data['year'] -
                                 1][data['month']] = data['last_year_revenue']
            except:
                pass

            if (data['month'] == this_month):
                monthly_dict[
                    data['year']][f"grand_total"] = data['grand_total']

                single_month_revenue_dict['revenue'] = data["revenue"]
                single_month_revenue_dict['YoY'] = data[
                    "revenue_increment_ratio"]

                grand_total_dict['revenue'] = data["grand_total"]
                grand_total_dict['YoY'] = data['grand_total_increment_ratio']

                revenue_by_year[
                    data['year']] = single_month_revenue_dict.copy()
                grand_total_by_year[data['year']] = grand_total_dict.copy()

        query_data['month_revenue'] = pd.DataFrame(monthly_dict)
        query_data['this_month_revenue_over_years'] = pd.DataFrame(
            revenue_by_year)
        query_data['grand_total_over_years'] = pd.DataFrame(
            grand_total_by_year)

        query_data.pop("monthly_data")

        return query_data

    def _expand_value_percentage(self, dataframe):

        expanded_columns = {}
        for col in dataframe.columns:
            # Use json_normalize to split 'value' and 'percentage'
            expanded_df = pd.json_normalize(
                dataframe[col]).add_prefix(f"{col}_")
            expanded_df.index = dataframe.index
            # Append the expanded columns to the new DataFrame
            expanded_columns[col] = expanded_df

        expanded_df = pd.concat(expanded_columns.values(), axis=1)

        return expanded_df

    def _get_today(self):
        today = datetime.today()
        this_year = today.year
        this_month = today.month
        this_day = today.day

        return {
            "today": today,
            "year": this_year,
            "month": this_month,
            "day": this_day,
        }

    def get_balance_sheet(self, ticker):
        """
        iFa.ai: 財務分析 -> 資產負債表

        Return: Dict
        {
            'ticker': 股票代碼,
            'company_name': 公司名稱,

            'balance_sheet': 歷年當季資場負債表"全表" (pd.DataFrame)
            'total_asset': 歷年當季資產總額 (pd.DataFrame)
            'current_asset': 歷年當季流動資產總額 (pd.DataFrame)
            'non_current_asset': 歷年當季非流動資產 (pd.DataFrame)
            'current_debt': 歷年當季流動負債 (pd.DataFrame)
            'non_current_debt': 歷年當季非流動負債 (pd.DataFrame)
            'equity': : 歷年當季權益 (pd.DataFrame)
        }
        """
        today_dict = self._get_today()

        today = today_dict['today']
        target_season = ((today.month - 1) // 3) + 1

        start_date = "2014-01-01"
        end_date = today.strftime("%Y-%m-%d")

        query_data = self.query_seasonal_data(ticker, start_date, end_date,
                                              "balance_sheet", target_season=target_season)

        return_dict = {
            "ticker": query_data[0]['ticker'],
            "company_name": query_data[0]['company_name'],
        }

        index_names = []

        table_dict = dict()
        total_asset_dict = dict()
        current_asset_dict = dict()
        non_current_asset_dict = dict()
        current_debt_dict = dict()
        non_current_debt_dict = dict()
        equity_dict = dict()

        this_season = query_data[-1]['seasonal_data']['season']

        value_type_list = ['value', 'percentage']

        for data in query_data:
            year = data['seasonal_data']['year']
            season = data['seasonal_data']['season']

            time_index = f"{year}Q{season}"

            if (season == this_season):
                try:
                    table_dict[time_index] = data['seasonal_data'][
                        'balance_sheet']
                except:
                    table_dict[time_index] = dict()
                    table_dict[time_index] = data['seasonal_data'][
                        'balance_sheet']

                try:
                    total_asset_dict[time_index] = {
                        "total_asset":
                        data['seasonal_data']['balance_sheet']['資產總額'],
                        "total_debt":
                        data['seasonal_data']['balance_sheet']['負債總額'],
                        "total_equity":
                        data['seasonal_data']['balance_sheet']['權益總額'],
                    }
                except:
                    total_asset_dict[time_index] = {
                        "total_asset": None,
                        "total_debt": None,
                        "total_equity": None,
                    }

                for value_type in value_type_list:
                    try:
                        current_asset_dict[
                            f"{time_index}_{value_type}"] = data[
                                'seasonal_data']['balance_sheet']['流動資產合計'][
                                    value_type]
                    except:
                        if (time_index not in current_asset_dict.keys()):
                            current_asset_dict[
                                f"{time_index}_{value_type}"] = None

                    try:
                        non_current_asset_dict[
                            f"{time_index}_{value_type}"] = data[
                                'seasonal_data']['balance_sheet']['非流動資產合計'][
                                    value_type]
                    except:
                        non_current_asset_dict[
                            f"{time_index}_{value_type}"] = None

                    try:
                        current_debt_dict[f"{time_index}_{value_type}"] = data[
                            'seasonal_data']['balance_sheet']['流動負債合計'][
                                value_type]
                    except:
                        current_debt_dict[f"{time_index}_{value_type}"] = None

                    try:
                        non_current_debt_dict[
                            f"{time_index}_{value_type}"] = data[
                                'seasonal_data']['balance_sheet']['非流動負債合計'][
                                    value_type]
                    except:
                        non_current_debt_dict[
                            f"{time_index}_{value_type}"] = None

                    try:
                        equity_dict[f"{time_index}_{value_type}"] = data[
                            'seasonal_data']['balance_sheet']['權益合計'][
                                value_type]
                    except:
                        equity_dict[f"{time_index}_{value_type}"] = None

                index_names += list(
                    data['seasonal_data']['balance_sheet'].keys())

        index_names = list(dict.fromkeys(index_names))

        balance_sheet_table = pd.DataFrame(table_dict)
        balance_sheet_table = self._expand_value_percentage(
            balance_sheet_table)

        total_asset_table = pd.DataFrame(total_asset_dict)
        total_asset_table = self._expand_value_percentage(total_asset_table)

        current_asset_table = pd.DataFrame(current_asset_dict,
                                           index=['current_asset'])
        non_current_asset_table = pd.DataFrame(non_current_asset_dict,
                                               index=['non_current_asset'])
        current_debt_table = pd.DataFrame(non_current_asset_dict,
                                          index=['current_debt'])
        non_current_debt_table = pd.DataFrame(non_current_asset_dict,
                                              index=['non_current_debt'])
        equity_table = pd.DataFrame(non_current_asset_dict, index=['equity'])

        return_dict['balance_sheet'] = balance_sheet_table
        return_dict['total_asset'] = total_asset_table
        return_dict['current_asset'] = current_asset_table
        return_dict['non_current_asset'] = non_current_asset_table
        return_dict['current_debt'] = current_debt_table
        return_dict['non_current_debt'] = non_current_debt_table
        return_dict['equity'] = equity_table
        return return_dict

    def _gen_dict(self,
                  query_data,
                  target_season,
                  keys,
                  calculate_type='value',
                  calculate_grand_total=False):
        """
        Will be deprecated
        """
        assert(calculate_type in ['growth_rate', 'value', 'percentage']), "args: calculate_type Error"
        table_dict = dict()
        grand_total_dict = dict() if (calculate_grand_total) else None

        for data in query_data:
            if (calculate_grand_total
                    and data['seasonal_data']['season'] <= target_season):
                time_index = f"{data['seasonal_data']['year']}Q{target_season}"
                profit_lose = data['seasonal_data']['profit_lose']

                for key in keys:
                    try:
                        if (calculate_type in ['growth_rate']):
                            for growth_rate in ['YoY_1', 'YoY_3', 'YoY_5', 'YoY_10']:
                                try:
                                    grand_total_dict[time_index][
                                        growth_rate] += profit_lose[key][growth_rate]
                                except Exception:
                                    if (time_index not in 
                                    grand_total_dict.keys()):
                                        grand_total_dict[time_index] = {
                                            "YoY": None,
                                            "YoY_3": None,
                                            "YoY_5": None,
                                            "YoY_10": None,
                                        }
                                    grand_total_dict[time_index][
                                        growth_rate] = profit_lose[key][growth_rate]

                        elif (calculate_type in ['percentage']):
                            grand_total_dict[time_index] += profit_lose[key][
                                calculate_type] / target_season
                        else:
                            grand_total_dict[time_index] += profit_lose[key][
                                calculate_type]
                        break
                    except KeyError:
                        try:
                            if (calculate_type
                                    in ['percentage']):
                                grand_total_dict[time_index] = profit_lose[
                                    key][calculate_type] / target_season
                            else:
                                grand_total_dict[time_index] = profit_lose[
                                    key][calculate_type]
                            break
                        except:  # key in profit_lose not found or not growth_rate not implemented
                            continue
                    except Exception:  # Other exceotion
                        continue
                else:  # All keys not found
                    grand_total_dict[time_index] = None

            if (data['seasonal_data']['season'] == target_season):
                time_index = f"{data['seasonal_data']['year']}Q{target_season}"
                profit_lose = data['seasonal_data']['profit_lose']

                for key in keys:
                    try:
                        if (calculate_type in ['growth_rate']):
                            for item in items:
                                table_dict[time_index][item] = profit_dict[key][item]
                        else:
                            table_dict[time_index] = profit_lose[key][
                            calculate_type]
                        break
                    except Exception:
                        continue
                else:
                    if (calculate_type == 'growth_rate'):
                        table_dict[time_index] = {
                            "YoY_1": None,
                            "YoY_3": None,
                            "YoY_5": None,
                            "YoY_10": None
                        }
                    else:
                        table_dict[time_index] = None
        return table_dict, grand_total_dict
    
    def _slice_multi_col_table(
        self,
        total_table,
        mode='value',
        target_index=None, # None or Str， 要特別抓哪個index
    ):
        times = total_table.columns.get_level_values(0).unique()
        try:
            target_metrics = self.target_metric_dict[mode]
        except KeyError as e:
            return f"mode Error: Get mode should be {list(self.target_metric_dict.keys())} but get {mode}"
        
        desired_order = [
            (time, value_name) for time in times for value_name in target_metrics
        ]

        if (target_index):
            sliced_table = total_table.loc[[target_index], pd.IndexSlice[:, target_metrics]][desired_order].T
            sliced_table = sliced_table.reset_index()
            sliced_table = sliced_table.pivot(index='level_1', columns='level_0', values=target_index)
            sliced_table.columns.name = None
            sliced_table.index.name = None
            return sliced_table.reindex(target_metrics)
        
        else:
            return total_table.loc[:, pd.IndexSlice[:, target_metrics]][desired_order]

    def get_profit_lose(self, ticker):
        """
        ticker: str
        iFa.ai: 財務分析 -> 損益表
        """
        today_dict = self._get_today()

        table_settings = StatsProcessor.load_yaml("profit_lose.yaml")
        today = today_dict['today'].strftime("%Y-%m-%d")
        start_date = "2014-01-01"

        this_season = ((today_dict['month'] - 1) // 3)
        this_season = 4 if (this_season == 0) else this_season - 1
        # TODO: 將這裡改成根據每公司的最後更新季度

        query_data = self.query_seasonal_data(ticker,
                                              start_date=start_date,
                                              end_date=today,
                                              sheet='profit_lose',
                                              target_season=this_season)

        index_names = []

        return_dict = {
            "ticker": query_data[0]['ticker'],
            "company_name": query_data[0]['company_name'],
        }

        table_dict = dict()
        grand_total_dict = dict()

        column_names = []

        for data in query_data:
            year = data['seasonal_data']['year']
            season = data['seasonal_data']['season']

            time_index = f"{year}Q{season}"

            index_names += list(
                data['seasonal_data']['profit_lose'].keys())

            profit_lose = data['seasonal_data']['profit_lose']
            
            for index_name, value_dict in profit_lose.items():
                column_names += [
                    (time_index, index_name, item_name) 
                    for item_name in value_dict.keys()
                ]
                for item_name, item in value_dict.items():
                    try:
                        table_dict[index_name][(time_index, item_name)] = item 
                        #[time_index][index_name][item_name] = item

                    except KeyError:
                        if (index_name not in table_dict.keys()):
                            table_dict[index_name] = dict()
                            grand_total_dict[index_name] = dict()
                        
                        table_dict[index_name][(time_index, item_name)] = item 

        columns = pd.MultiIndex.from_tuples(table_dict.keys())
        total_table = pd.DataFrame.from_dict(
            table_dict,
            orient='index'
        )

        total_table.columns = pd.MultiIndex.from_tuples(total_table.columns)

        for name, setting in table_settings.items():
            return_dict[name] = self._slice_multi_col_table(
                total_table=total_table,
                mode=setting['mode'],
                target_index=setting['target_index'] if "target_index" in setting.keys() else None 
            )
        
        return return_dict

    def get_cash_flow(self, ticker):
        """
        iFa.ai: 財務分析 -> 現金金流表
        """
        today_dict = self._get_today()

        today = today_dict['today']
        this_season = (today.month - 1) // 3 + 1
        start_date = "2014-01-01"
        end_date = today.strftime("%Y-%m-%d")

        query_data = self.query_seasonal_data(ticker,
                                              start_date=start_date,
                                              end_date=end_date,
                                              sheet='cash_flow',
                                              target_season=this_season)

        index_names = []

        return_dict = {
            "ticker": query_data[0]['ticker'],
            "company_name": query_data[0]['company_name'],
        }

        table_dict = dict()
        CASHO_dict = dict()  # 營業活動
        CASHI_dict = dict()  # 投資活動
        CASHF_dict = dict()  # 籌資活動

        this_season = query_data[-1]['seasonal_data']['season']

        checkpoints = ["營業活動之現金流量－間接法", "投資活動之現金流量", "籌資活動之現金流量"]
        main_cash_flows = [
            "營業活動之淨現金流入（流出）", "投資活動之淨現金流入（流出）", "籌資活動之淨現金流入（流出）"
        ]

        partial_cash_flows = [CASHO_dict, CASHI_dict, CASHF_dict]

        for data in query_data:
            year = data['seasonal_data']['year']
            season = data['seasonal_data']['season']

            time_index = f"{year}Q{season}"

            if (season == this_season):
                cash_flow = data['seasonal_data']['cash_flow']
                main_cash_flow_name = None
                partial_cash_flow = None
                next_checkpoint = 0

                for index_name, value in cash_flow.items():
                    if (next_checkpoint < 3
                            and index_name == checkpoints[next_checkpoint]):
                        main_cash_flow_name = main_cash_flows[next_checkpoint]
                        partial_cash_flow = partial_cash_flows[next_checkpoint]
                        next_checkpoint += 1
                    try:
                        table_dict[time_index][index_name]['value'] = value[
                            'value']
                        if (value['value']):
                            table_dict[time_index][index_name][
                                'percentage'] = value['value'] / cash_flow[
                                    main_cash_flow_name]['value']
                        else:
                            table_dict[time_index][index_name][
                                'percentage'] = None
                    except:
                        if (time_index not in table_dict.keys()):
                            table_dict[time_index] = dict()
                        table_dict[time_index][index_name] = dict()

                        table_dict[time_index][index_name]['value'] = value[
                            'value']
                        if (value['value']):
                            table_dict[time_index][index_name][
                                'percentage'] = value['value'] / cash_flow[
                                    main_cash_flow_name]['value']
                        else:
                            table_dict[time_index][index_name][
                                'percentage'] = None

                    try:
                        partial_cash_flow[time_index][index_name] = table_dict[
                            time_index][index_name]
                    except:
                        if (time_index not in partial_cash_flow.keys()):
                            partial_cash_flow[time_index] = dict()
                        partial_cash_flow[time_index][index_name] = table_dict[
                            time_index][index_name]

                index_names += list(cash_flow.keys())

        index_names = list(dict.fromkeys(index_names))

        cash_flow_table = pd.DataFrame(table_dict)
        cash_flow_table = self._expand_value_percentage(cash_flow_table)

        CASHO_table = pd.DataFrame(CASHO_dict)
        CASHO_table = self._expand_value_percentage(CASHO_table)

        CASHI_table = pd.DataFrame(CASHI_dict)
        CASHI_table = self._expand_value_percentage(CASHI_table)

        CASHF_table = pd.DataFrame(CASHF_dict)
        CASHF_table = self._expand_value_percentage(CASHF_table)

        return_dict['cash_flow'] = cash_flow_table
        return_dict['CASHO'] = CASHO_table
        return_dict['CASHI'] = CASHI_table
        return_dict['CASHF'] = CASHF_table

        return return_dict


from .base import StatsFetcher, StatsDateTime
import json
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor, YoY_Calculator
import importlib.resources as pkg_resources
import yaml


class MonthRevenueFetcher(StatsFetcher):
    """
        iFa.ai: 財務分析 -> 每月營收
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

    def prepare_query(self, target_year, target_month):
        pipeline = super().prepare_query()

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "monthly_data": {
                    "$sortArray": {
                        "input": "$monthly_data",
                        "sortBy": {
                            "year": -1,
                            "month": -1
                        }
                    }
                },
            }
        })

        return pipeline

    def collect_data(self, target_year, target_month):
        pipeline = self.prepare_query(target_year, target_month)

        fetched_data = self.collection.aggregate(pipeline)

        fetched_data = list(fetched_data)

        return fetched_data[-1]

    def query_data(self):
        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection)['last_update_time']
            target_year = latest_time['monthly_data']['latest_year']
            target_month = latest_time['monthly_data']['latest_month']
        except Exception as e:
            today = StatsDateTime.get_today()
            target_month = today.month
            target_year = today.year

        # Query data
        fetched_data = self.collect_data(target_year, target_month)

        return self.process_data(fetched_data)

    def process_data(self, fetched_data):

        monthly_data = fetched_data['monthly_data']
        for data in monthly_data:
            for key, value in data.items():
                if ("YoY" in key):
                    data[key] = StatsProcessor.cal_percentage(value)
                elif ("ratio" in key or 'percentage' in key):
                    data[key] = StatsProcessor.cal_non_percentage(value,
                                                                  to_str=True,
                                                                  postfix="%")
                elif (key not in ('year', 'month')):
                    data[key] = StatsProcessor.cal_non_percentage(value,
                                                                  postfix="千元")
        target_month = monthly_data[0]['month']
        monthly_df = pd.DataFrame(monthly_data)

        target_month_df = monthly_df[monthly_df['month'] == target_month]
        annual_month_df = monthly_df[monthly_df['month'] == 12]
        month_revenue_df = monthly_df.pivot(index='month',
                                            columns='year',
                                            values='revenue')

        grand_total_df = target_month_df.pivot(index='month',
                                               columns='year',
                                               values='grand_total')

        annual_total_df = annual_month_df.pivot(index='month',
                                               columns='year',
                                               values='grand_total')

        grand_total_df.rename(index={target_month: f"grand_total"},
                              inplace=True)
        month_revenue_df = month_revenue_df.sort_index(ascending=False)
        month_revenue_df = pd.concat([grand_total_df, month_revenue_df],
                                     axis=0)

        fetched_data['month_revenue'] = month_revenue_df[sorted(
            month_revenue_df.columns, reverse=True)]
        # 歷年月營收
        fetched_data[
            'this_month_revenue_over_years'] = target_month_df.set_index(
                "year")[[
                    "revenue", "revenue_increment_ratio", "YoY_1", "YoY_3",
                    "YoY_5", "YoY_10"
                ]].T
        # 歷年營收成長量
        fetched_data['grand_total_over_years'] = target_month_df.set_index(
            "year")[[
                "grand_total", "grand_total_increment_ratio",
                "grand_total_YoY_1", "grand_total_YoY_3", "grand_total_YoY_5",
                "grand_total_YoY_10"
            ]].T

        fetched_data.pop("monthly_data")

        fetched_data['recent_month_revenue'] = self.get_recent_revenue_grwoth(
            monthly_data, grand_total_dict=annual_total_df.to_dict(), interval = 12
        )

        return fetched_data

    def get_recent_revenue_grwoth(self, monthly_data, grand_total_dict, interval: int = 12):
        recent_month_data = monthly_data[:interval + 1]

        MoMs = [
            YoY_Calculator.cal_growth(this_value['revenue'], last_value['revenue'], delta = 1)
            for this_value, last_value in zip(
                recent_month_data[:12], recent_month_data[1:13]
            )
        ]

        recent_month_data = {
            "date" : [f"{data['year']}/{data['month']}" for data in recent_month_data[:interval]],
            "revenue" : [data['revenue'] for data in recent_month_data[:interval]],
            "MoM" : [f"{(data * 100):.2f}%" for data in MoMs],
            "YoY" : [f"{data['revenue_increment_ratio']}" for data in recent_month_data[:interval]],
            "total_YoY":  [f"{data['grand_total_increment_ratio']}" for data in recent_month_data[:interval]],
        }

        # accum_YoY
        # accum_YoY 為 Davis提出的定義
        # 2024/6的累計YoY(accum_YoY) 為 2024累計到6月為止的總營收/2023年度總營收
        accum_YoYs = []
        for data in monthly_data[:interval]:
            try:
                year = data['year'] - 1
                total = grand_total_dict[year][12]
                accum_YoY = round((data['grand_total'] / total) * 100, 2)
                accum_YoYs.append(f"{accum_YoY}%")
            except Exception as e:
                accum_YoYs.append(None)
        
        recent_month_data['accum_YoY'] = accum_YoYs

        recent_month_df = pd.DataFrame(recent_month_data).set_index('date').T

        return recent_month_df


import pandas as pd

from utils import StatsFetcher

pd.options.display.max_rows = 4
pd.options.display.max_columns = 4

if __name__ == "__main__":
    import pprint
    pp = pprint.PrettyPrinter(indent=2)

    ticker = "2330"
    start_date = "2018-01-01"
    end_date = "2199-12-31"

    fetcher = StatsFetcher()

    # pp.pprint(fetcher.query_seasonal_data(ticker, start_date, end_date, 'balance_sheet'))
    # fetcher.get_balance_sheet(ticker)
    query_data = fetcher.get_profit_lose(ticker)
    print("{")
    for key, value in query_data.items():
        print(f"\t\"{key}\":")

        print(f"{value}")
    print("}")
    # pp.pprint(fetcher.get_balance_sheet(ticker))


from .base import StatsFetcher, StatsDateTime
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import importlib.resources as pkg_resources
import yaml

class CashFlowFetcher(StatsFetcher):
    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.cash_flow_dict = StatsProcessor.load_yaml(
            "cash_flow_percentage.yaml"
        )  # 計算子表格用
    
    def prepare_query(self, target_season):
        pipeline = super().prepare_query()

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "cash_flows": {
                    "$sortArray": {
                        "input": {
                            "$map": {
                                "input": {
                                    "$filter": {
                                        "input": "$seasonal_data",
                                        "as": "season",
                                        "cond": {
                                            "$eq":
                                            ["$$season.season", target_season]
                                        }
                                    }
                                },
                                "as": "target_season_data",
                                "in": {
                                    "year":
                                    "$$target_season_data.year",
                                    "season":
                                    "$$target_season_data.season",
                                    "cash_flow":
                                    "$$target_season_data.cash_flow"
                                }
                            }
                        },
                        "sortBy": {
                            "year": -1
                        }  # 按 year 降序排序
                    }
                }
            }
        })

        return pipeline
    
    def collect_data(self, target_season):
        pipeline = self.prepare_query(target_season)

        fetched_data = self.collection.aggregate(pipeline)

        return list(fetched_data)[0]
    
    def query_data(self):
        
        try:
            latest_time = StatsDateTime.get_latest_time(self.ticker, self.collection)['last_update_time']
            target_season = latest_time['seasonal_data']['latest_season']
        
        except:
            today = StatsDateTime.get_today()
            target_season = today.season - 1 if (today.season > 1) else 4

        fetched_data = self.collect_data(target_season)

        return self.process_data(fetched_data, target_season)
    
    def process_data(self, fetched_data, target_season):
        """
        處理現金流量表頁面的所有表格
        金流表本身沒有比例 但是Ifa有算，
        項目所屬的情況也不一(分別所屬營業,投資,籌資三個活動)
        所以這裡選擇不用slicing處理
        """
        cash_flows = fetched_data['cash_flows']

        index_names = []
        column_names = []

        table_dict = dict()
        CASHO_dict = dict()
        CASHI_dict = dict()
        CASHF_dict = dict()

        return_dict = {
            "ticker": fetched_data['ticker'],
            "company_name": fetched_data['company_name'],
            "cash_flow": dict(),
            "CASHO": dict(),
            "CASHI": dict(),
            "CASHF": dict()
        }

        checkpoints = ["營業活動之現金流量－間接法", "投資活動之現金流量", "籌資活動之現金流量", "匯率變動對現金及約當現金之影響"]
        main_cash_flows = [
            "營業活動之淨現金流入（流出）", "投資活動之淨現金流入（流出）", "籌資活動之淨現金流入（流出）", None
        ] # 主要的比例對象
        partial_cash_flows = [CASHO_dict, CASHI_dict, CASHF_dict, dict()] 

        # 作法: dictionary中也有checkpoints，如果出現了就換下一個index去計算

        for data in cash_flows:
            year = data['year']

            time_index = f"{year}Q{target_season}"

            cash_flow = data['cash_flow']
            main_cash_flow_name = None
            partial_cash_flow = None
            next_checkpoint = 0

            for index_name, value in cash_flow.items():
                if (next_checkpoint < 3
                        and index_name == checkpoints[next_checkpoint]): # 找到了主要的變動點
                    main_cash_flow_name = main_cash_flows[next_checkpoint]
                    partial_cash_flow = partial_cash_flows[next_checkpoint]
                    next_checkpoint += 1
                try:
                    table_dict[time_index][index_name]['value'] = value[
                        'value']
                    if (value['value']):
                        ratio = np.round(
                                (value['value'] / cash_flow[
                                main_cash_flow_name]['value']) * 100, 2)
                        table_dict[time_index][index_name][
                            'percentage'] = f"{ratio}%"
                    else:
                        table_dict[time_index][index_name][
                            'percentage'] = None
                except: # 新增index再做一次
                    if (time_index not in table_dict.keys()):
                        table_dict[time_index] = dict()
                    table_dict[time_index][index_name] = dict()

                    table_dict[time_index][index_name]['value'] = value[
                        'value']
                    if (value['value']):
                        ratio = np.round(
                                (value['value'] / cash_flow[
                                main_cash_flow_name]['value']) * 100, 2) 
                        table_dict[time_index][index_name][
                            'percentage'] = f"{ratio}%"
                    else:
                        table_dict[time_index][index_name][
                            'percentage'] = None
                table_dict[time_index][index_name]['value'] = StatsProcessor.cal_non_percentage(value['value'], postfix="千元")
                try:
                    partial_cash_flow[time_index][index_name] = table_dict[
                        time_index][index_name]
                except:
                    if (time_index not in partial_cash_flow.keys()):
                        partial_cash_flow[time_index] = dict()
                    partial_cash_flow[time_index][index_name] = table_dict[
                        time_index][index_name]

            index_names += list(cash_flow.keys())

        # 轉成dictionary keys
        index_names = list(dict.fromkeys(index_names))

        cash_flow_table = pd.DataFrame(table_dict)
        cash_flow_table = StatsProcessor.expand_value_percentage(cash_flow_table)

        CASHO_table = pd.DataFrame(CASHO_dict)
        CASHO_table = StatsProcessor.expand_value_percentage(CASHO_table)

        CASHI_table = pd.DataFrame(CASHI_dict)
        CASHI_table = StatsProcessor.expand_value_percentage(CASHI_table)

        CASHF_table = pd.DataFrame(CASHF_dict)
        CASHF_table = StatsProcessor.expand_value_percentage(CASHF_table)

        return_dict['cash_flow'] = cash_flow_table
        return_dict['CASHO'] = CASHO_table
        return_dict['CASHI'] = CASHI_table
        return_dict['CASHF'] = CASHF_table

        return return_dict



from .base import BaseTEJFetcher
from datetime import datetime
from enum import Enum
import pandas as pd
from pymongo import MongoClient
from .tech import TechProcessor
from ..utils import StatsProcessor, YoY_Calculator
import warnings
import yaml


class FinanceReportFetcher(BaseTEJFetcher):

    class FetchMode(Enum):
        YOY = 1
        QOQ = 2
        YOY_NOCAL = 3
        QOQ_NOCAL = 4

    def __init__(
            self,
            mongo_uri,
            db_name="company",
            collection_name="TWN/AINVFQ1"
        ):
        self.client = MongoClient(mongo_uri)
        self.db = self.client[db_name]
        self.collection = self.db[collection_name]

        index_dict = StatsProcessor.load_yaml("tej_db/tej_db_index.yaml")
        thousand_dict = StatsProcessor.load_yaml("tej_db/tej_db_thousand_index.yaml")
        percent_dict = StatsProcessor.load_yaml("tej_db/tej_db_percent_index.yaml")
        skip_dict = StatsProcessor.load_yaml("tej_db/tej_db_percent_index.yaml")
        self.check_index = set(index_dict[collection_name])
        self.skip_index = set(skip_dict[collection_name])

        self.thousand_index_list = list(thousand_dict[collection_name])
        self.percent_index_list = list(percent_dict[collection_name])


    def get(
            self,
            ticker,
            fetch_mode: FetchMode = FetchMode.QOQ_NOCAL,
            start_date: str = None,
            end_date: str = None,
            report_type: str = "Q",
            indexes: list = []):
        """
        基礎的query function
        ticker(str): 股票代碼
        start_date(str):  開頭日期範圍
        end_date(str):  = 結束日期範圍
        report_type(str): 報告型態 {"A", "Q", "TTM"}  
        fetch_mode(class FetchMode): 
           YoY : 起始日期到結束日期範圍內，特定該季的資料
           QoQ : 起始日期到結束日期內，每季的資料(與上一季成長率)
        indexes(List): 指定的index
        """
        # 確認indexes中是否有錯誤的index，有的話回傳warning
        if (indexes and self.check_index):
            indexes = set(indexes)
            difference = indexes - self.check_index
            if (difference):
                warnings.warn(
                    f"{list(difference)} 沒有出現在資料表中，請確認column名稱是否正確",
                    UserWarning)

        if (not start_date):
            start_date = datetime.strptime("2005-01-01", "%Y-%m-%d")
        else:
            start_date = datetime.strptime(start_date, "%Y-%m-%d")

        if (fetch_mode in {self.FetchMode.QOQ, self.FetchMode.QOQ_NOCAL}):

            if (not end_date):
                end_date = datetime.today()
            else:
                end_date = datetime.strptime(end_date, "%Y-%m-%d")

            assert (start_date <= end_date)

            start_year = start_date.year
            start_season = (start_date.month - 1) // 4 + 1
            end_year = end_date.year
            end_season = (end_date.month - 1) // 4 + 1

            if (fetch_mode == self.FetchMode.QOQ):
                use_cal = True
            else:
                use_cal = False

            data_df = self.get_QoQ_data(
                ticker=ticker,
                start_year=start_year,
                start_season=start_season,
                end_year=end_year,
                end_season=end_season,
                report_type=report_type,
                indexes=indexes,
                use_cal=use_cal)

            return data_df

        elif (fetch_mode in {self.FetchMode.YOY, self.FetchMode.YOY_NOCAL}):
            start_year = start_date.year
            end_date = self.get_latest_data_time(ticker)
            if (not end_date):
                end_date = datetime.today()

            end_year = end_date.year
            season = (end_date.month - 1) // 4 + 1

            if (fetch_mode == self.FetchMode.YOY):
                use_cal = True
            else:
                use_cal = False

            data_df = self.get_YoY_data(
                ticker=ticker,
                start_year=start_year,
                end_year=end_year,
                season=season,
                report_type=report_type,
                indexes=indexes,
                use_cal=use_cal)

            return data_df

    def get_QoQ_data(
            self,
            ticker,
            start_year,
            start_season,
            end_year,
            end_season,
            report_type="Q",
            indexes=[],
            use_cal=False):
        """
            取得時間範圍內每季資料
        """
        if (use_cal):
            if (start_season == 1):
                lower_bound_year = start_year - 1
                lower_bound_season = 4

            else:
                lower_bound_year = start_year
                lower_bound_season = start_season - 1

        else:
            lower_bound_year = start_year,
            lower_bound_season = start_season

        if (not indexes):    # 沒有指定 -> 取全部
            pipeline = [
                {
                    "$match": {
                        "ticker": ticker
                    }
                }, {
                    "$unwind": "$data"
                }, {
                    "$match": {
                        "$or": [
                            {
                                "data.year": {
                                    "$gt": start_year,
                                    "$lt": end_year
                                }
                            }, {
                                "data.year": start_year,
                                "data.season": {
                                    "$gte": start_season
                                }
                            }, {
                                "data.year": end_year,
                                "data.season": {
                                    "$lte": end_season
                                }
                            }, {
                                "data.year": lower_bound_year,
                                "data.season": lower_bound_season
                            }
                        ]
                    }
                }, {
                    "$project": {
                        "data.year": 1,
                        "data.season": 1,
                        f"data.{report_type}": 1,
                        "_id": 0
                    }
                }
            ]

        else:    # 取指定index
            project_stage = {"data.year": 1, "data.season": 1}
            for index in indexes:
                project_stage[f"data.{report_type}.{index}"] = 1

            pipeline = [
                {
                    "$match": {
                        "ticker": ticker
                    }
                }, {
                    "$unwind": "$data"
                }, {
                    "$match": {
                        "$or": [
                            {
                                "data.year": {
                                    "$gt": start_year,
                                    "$lt": end_year
                                }
                            }, {
                                "data.year": start_year,
                                "data.season": {
                                    "$gte": start_season
                                }
                            }, {
                                "data.year": end_year,
                                "data.season": {
                                    "$lte": end_season
                                }
                            }, {
                                "data.year": lower_bound_year,
                                "data.season": lower_bound_season
                            }
                        ]
                    }
                }, {
                    "$project": project_stage
                }
            ]

        fetched_data = self.collection.aggregate(pipeline).to_list()
        data_dict = StatsProcessor.list_of_dict_to_dict(
            fetched_data,
            keys=["year", "season"],
            delimeter="Q",
            data_key=report_type)
        
        data_dict = self.transform_value(data_dict)

        if (use_cal):
            data_with_QoQ = self.cal_QoQ(data_dict)
            data_df = pd.DataFrame.from_dict(data_with_QoQ)
            data_df = data_df.iloc[:, 1:]
            data_df = data_df.iloc[:, ::-1].T
            data_dict = data_df.to_dict()
            data_dict = self.get_dict_of_df(data_dict)
            return data_dict
        else:
            data_df = pd.DataFrame.from_dict(data_dict)
            data_df = data_df.iloc[:, ::-1]
            return data_df

    def get_YoY_data(
            self,
            ticker,
            start_year,
            end_year,
            season,
            report_type="Q",
            indexes=[],
            use_cal=False):
        """
        取得某季歷年資料
        """
        if (use_cal):
            select_year = set()

            for year in range(start_year, end_year + 1):
                year_shifts = {year, year - 1, year - 3, year - 5, year - 10}

                select_year = select_year.union(year_shifts)

            select_year = sorted(list(select_year), reverse=True)
        else:
            select_year = [year for year in range(start_year, end_year + 1)]

        if (not indexes):    # 沒有指定 -> 取全部
            pipeline = [
                {
                    "$match": {
                        "ticker": ticker
                    }
                }, {
                    "$unwind": "$data"
                }, {
                    "$match": {
                        "$or": [
                            {
                                "$and": [
                                    {
                                        "data.year": {
                                            "$in": select_year
                                        }
                                    }, {
                                        "data.season": {
                                            "$eq": season
                                        }
                                    }
                                ]
                            },
                        ]
                    }
                }, {
                    "$project": {
                        "data.year": 1,
                        "data.season": 1,
                        f"data.{report_type}": 1,
                        "_id": 0
                    }
                }
            ]

        else:    # 取指定index
            project_stage = {"data.year": 1, "data.season": 1}
            for index in indexes:
                project_stage[f"data.{report_type}.{index}"] = 1

            pipeline = [
                {
                    "$match": {
                        "ticker": ticker
                    }
                }, {
                    "$unwind": "$data"
                }, {
                    "$match": {
                        "$and": [
                            {
                                "data.year": {
                                    "$in": select_year
                                }
                            }, {
                                "data.season": {
                                    "$eq": season
                                }
                            }
                        ]
                    }
                }, {
                    "$project": project_stage
                }
            ]

        fetched_data = self.collection.aggregate(pipeline).to_list()

        # 處理計算YoY
        data_dict = StatsProcessor.list_of_dict_to_dict(
            fetched_data,
            keys=['year', 'season'],
            data_key=report_type,
            delimeter='Q')
        
        data_dict = self.transform_value(data_dict)

        if (use_cal):
            data_with_YoY = self.cal_YoY(
                data_dict, start_year, end_year, season)
            data_df = pd.DataFrame.from_dict(data_with_YoY)
            data_df = data_df.iloc[:, ::-1].T
            data_dict = data_df.to_dict()
            data_dict = self.get_dict_of_df(data_dict)
            return data_dict
        else:
            data_df = pd.DataFrame.from_dict(data_dict)
            data_df = data_df.iloc[:, ::-1]
            return data_df
        
    def transform_value(self, data_dict):
        """
        處理千元, %等單位
        """

        data_df = pd.DataFrame.from_dict(data_dict)
        
        process_set = set(data_df.index).intersection(set(self.thousand_index_list))
        process_list = list(process_set)
        data_df.loc[process_list] = data_df.loc[process_list].map(
            lambda x : StatsProcessor.cal_non_percentage(x, postfix="千元")
        )

        process_set = set(data_df.index).intersection(set(self.percent_index_list))
        process_list = list(process_set)
        data_df.loc[process_list] = data_df.loc[process_list].map(
            lambda x : f"{x}%"
        )

        data_dict = data_df.to_dict()
        
        return data_dict

class TEJStockPriceFetcher(BaseTEJFetcher):

    def __init__(
            self,
            mongo_uri,
            db_name: str = "company",
            collection_name: str = None):
        self.mongo_uri = mongo_uri
        self.db_name = db_name
        self.collection_name = collection_name

        self.client = MongoClient(self.mongo_uri)
        self.db = self.client[self.db_name]
        self.collection = self.db[self.collection_name]

        self.check_period = ['1d', '7d', '1m', '3m', '1y', '3y', '5y', '10y', 'all']

    def get(
            self,
            ticker: str = "2330",
            start_date: str = "2024-10-01",
            period: str = None
        ):
        """
        取得開高低收資料
        start_date: str: 起始的日期
        period: 指定日期範圍(E.g. 1天, 7天...etc)
        如果宣告period, 以period為優先
        """

        assert (
            period is None or period in self.check_period
        ), f"period should be None or {','.join(self.check_period)}"

        if (period is not None):
            latest_date = self.get_latest_data_time(ticker)
            start_date = self.set_time_shift(date=latest_date, period=period)
        else:
            start_date = datetime.strptime(start_date, "%Y-%m-%d")

        pipeline = [
            {
                "$match": {
                    "ticker": ticker
                }
            }, {
                "$unwind": "$data"
            }, {
                "$match": {
                    "data.mdate": {
                        "$gt": start_date
                    }
                }
            }, {
                "$project": {
                    "ticker": 1,
                    "data": 1,
                    "_id": 0
                }
            }
        ]
        datas = self.collection.aggregate(pipeline).to_list()

        elements = [element['data'] for element in datas]

        data_df = pd.DataFrame(elements).set_index('mdate')

        return data_df


import json
from pprint import PrettyPrinter
import pytest
import pandas as pd
import yaml
from neurostats_API.utils import StatsProcessor, DBClient

pd.set_option('display.max_rows', 5)
pd.set_option('display.max_columns', 4)

pp = PrettyPrinter(
    indent=2
)

with open(f"./neurostats_API/tw50_company_list.json") as f:
    company_list = json.load(f)
db_client = DBClient("mongodb://neurowatt:neurodb123@db.neurowatt.ai:27017/neurowatt").get_client()

def test_value_serie():
    from neurostats_API.fetchers import ValueFetcher
    company_list = {
        "2330": "台積電"
    }
    for ticker in company_list.keys():
    
        fetcher = ValueFetcher(ticker=ticker, db_client=db_client)

        fetched_data = fetcher.query_value_serie()

        pp.pprint(fetched_data)

def test_value():
    from neurostats_API.fetchers import ValueFetcher
    for ticker in company_list.keys():
    
        fetcher = ValueFetcher(ticker=ticker, db_client=db_client)

        fetched_data = fetcher.query_data()
    

    assert ('daily_data' in fetched_data)
    assert ('yearly_data' in fetched_data)

    assert isinstance(fetched_data['daily_data'], dict)
    assert isinstance(fetched_data['yearly_data'], pd.DataFrame)
    print("============================VALUE INVEST\n")
    for key, item in fetched_data.items():
        print(f"\"{key}\":\n{item}")
    

def test_profit_lose():
    from neurostats_API.fetchers import ProfitLoseFetcher
    for ticker in company_list.keys():
    
        fetcher = ProfitLoseFetcher(ticker, db_client)

        data = fetcher.query_data()

    table_settings = StatsProcessor.load_yaml("profit_lose.yaml")
    
    for key in table_settings.keys():
        assert key in data.keys()
    print("============================PROFIT LOSE\n")
    for key, item in data.items():
        print(f"\"{key}\":\n{item}")

def test_cash_flow():
    from neurostats_API.fetchers import CashFlowFetcher

    for ticker in company_list.keys():
        
        fetcher = CashFlowFetcher(ticker, db_client)
        data = fetcher.query_data()

    assert("cash_flow" in data.keys())
    assert("CASHO" in data.keys())
    assert("CASHI" in data.keys())
    assert("CASHF" in data.keys())

    print("============================CASH FLOW: \n")
    for key, item in data.items():
        print(f"\"{key}\":\n{item}")

def test_month_revenue():
    from neurostats_API.fetchers import MonthRevenueFetcher
    for ticker in company_list.keys():
    
        fetcher = MonthRevenueFetcher(ticker, db_client)
        data = fetcher.query_data()

    assert("month_revenue" in data.keys())
    assert("this_month_revenue_over_years" in data.keys())
    assert("grand_total_over_years" in data.keys())
    print("============================Month Revenue:\n")
    for key, item in data.items():
        print(f"\"{key}\":\n{item}")


def test_balance_sheet():
    from neurostats_API.fetchers import BalanceSheetFetcher
    for ticker in company_list.keys():
        
        fetcher = BalanceSheetFetcher(ticker=ticker, db_client=db_client)

        data = fetcher.query_data()
        
    assert("balance_sheet" in data.keys())
    assert('total_asset' in data.keys())
    assert('current_asset' in data.keys())
    assert('non_current_asset' in data.keys())
    assert('current_debt' in data.keys())
    assert('non_current_debt' in data.keys())
    assert('equity' in data.keys())
    print("============================BALANCE SHEET:\n")
    for key, item in data.items():
        print(f"\"{key}\":\n{item}")

def test_finance_overview():
    from neurostats_API.fetchers import FinanceOverviewFetcher
    company_list = {"2330": "台積電"}
    for ticker in company_list.keys():
        fetcher = FinanceOverviewFetcher(ticker=ticker, db_client=db_client)
        fetched_data = fetcher.query_data()

        # except Exception as e:
        #     print(f"{ticker} finance_overview {str(e)}")

    expected_keys = [
    # Queried items
        'revenue',
        'gross_profit',
        'operating_income',
        'net_income',
        'operating_cash_flow',
        'invest_cash_flow',
        'financing_cash_flow',
        'capital',
        'eps',
        'total_assets',
        'equity',
        'net_income_before_tax',
        'interest',
        'operating_expenses',
        'net_income_rate',
        'revenue_YoY',
        'gross_prof_YoY',
        'operating_income_YoY',
        'net_income_YoY',
        'account_receive',
        'account_pay',
        'inventories',
        'operating_cost',
        'application',
        'current_assets',
        'current_liabilities',
        'total_liabilities',
        'cash_and_cash_equivalents',
        'interest_expense',

        # calculated_items
        'fcf',
        'EBIT',
        'share_outstanding',
        'revenue_per_share',
        'gross_per_share',
        'operating_income_per_share',
        'operating_cash_flow_per_share',
        'fcf_per_share',
        'roa',
        'roe',
        'gross_over_asset',
        'roce',
        'gross_profit_margin',
        'operation_profit_rate',
        'operating_cash_flow_profit_rate',
        'dso',
        'account_receive_over_revenue',
        'dpo',
        'inventories_cycle_ratio',
        'dio',
        'inventories_revenue_ratio',
        'cash_of_conversion_cycle',
        'asset_turnover',
        'application_turnover',
        'current_ratio',
        'quick_ratio',
        'debt_to_equity_ratio',
        'net_debt_to_equity_ratio',
        'interest_coverage_ratio',
        'debt_to_operating_cash_flow',
        'debt_to_free_cash_flow',
        'cash_flow_ratio',
    ]

    for key in expected_keys:
        assert key in fetched_data['seasonal_data'], f"{key} not found in fetched_data"
        # assert fetched_data['seasonal_data'][0][key] is not None, f"{key} is None"

    pp.pprint(fetched_data)

def test_institution_trading():
    from neurostats_API.fetchers import InstitutionFetcher
    fetcher = InstitutionFetcher(ticker='2330', db_client = db_client)

    fetched_data = fetcher.query_data()

    pp.pprint(fetched_data)

def test_margin_trading():
    from neurostats_API.fetchers import MarginTradingFetcher
    fetcher = MarginTradingFetcher(ticker='2330', db_client=db_client)

    fetched_data = fetcher.query_data()

    pp.pprint(fetched_data)

def test_tech():
    from neurostats_API.fetchers import TechFetcher

    fetcher = TechFetcher(ticker = '1260', db_client=db_client)

    fetched_data = fetcher.get_quarterly()

    pp.pprint(fetched_data)


from .base import StatsDateTime, StatsFetcher
from .balance_sheet import BalanceSheetFetcher
from .cash_flow import CashFlowFetcher
from .finance_overview import FinanceOverviewFetcher
from .tej_finance_report import FinanceReportFetcher, TEJStockPriceFetcher
from .tech import TechFetcher
from .institution import InstitutionFetcher
from .margin_trading import MarginTradingFetcher
from .month_revenue import MonthRevenueFetcher
from .profit_lose import ProfitLoseFetcher
from .value_invest import ValueFetcher


from .base import StatsFetcher, StatsDateTime
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import importlib.resources as pkg_resources
import yaml


class BalanceSheetFetcher(StatsFetcher):
    """
    對應iFa.ai -> 財務分析 ->  資產負債表
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)
        self.table_settings = StatsProcessor.load_yaml("balance_sheet.yaml")

    def prepare_query(self, target_year, target_season):
        pipeline = super().prepare_query()

        target_query = {
            "year": "$$target_season_data.year",
            "season": "$$target_season_data.season",
            "balance_sheet": "$$$$target_season_data.balance_sheet"
        }

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "balance_sheets": {
                    "$sortArray": {
                        "input": {
                            "$map": {
                                "input": {
                                    "$filter": {
                                        "input": "$seasonal_data",
                                        "as": "season",
                                        "cond": {
                                            "$eq":
                                            ["$$season.season", target_season]
                                        }
                                    }
                                },
                                "as": "target_season_data",
                                "in": {
                                    "year":
                                    "$$target_season_data.year",
                                    "season":
                                    "$$target_season_data.season",
                                    "balance_sheet":
                                    "$$target_season_data.balance_sheet"
                                }
                            }
                        },
                        "sortBy": {
                            "year": -1
                        }  # 按 year 降序排序
                    }
                }
            }
        })

        return pipeline

    def collect_data(self, target_year, target_season):
        pipeline = self.prepare_query(target_year, target_season)

        fetched_data = self.collection.aggregate(pipeline)

        fetched_data = list(fetched_data)

        return fetched_data[-1]

    def query_data(self):
        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection)['last_update_time']
            year = latest_time['seasonal_data']['latest_year']
            season = latest_time['seasonal_data']['latest_season']
        except Exception as e:
            today = StatsDateTime.get_today()
            year = today.year - 1 if (today.season == 1) else today.year
            season = 4 if (today.season == 1) else today.season - 1

        fetched_data = self.collect_data(year, season)

        return self.process_data(season, fetched_data)

    def process_data(self, target_season, fetched_data):
        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name']
        }

        index_names = []

        table_dict = dict()

        balance_sheets = fetched_data['balance_sheets']

        # 將value與percentage跟著年分季度一筆筆取出
        for data in balance_sheets:
            year = data['year']

            time_index = f"{year}Q{target_season}"

            # 蒐集整體的keys
            index_names += list(data['balance_sheet'].keys())
            balance_sheet = data['balance_sheet']

            for index_name, value_dict in balance_sheet.items():
                for item_name, item in value_dict.items():
                    try:  # table_dict[項目][(2020Q1, '%')]
                        if (item_name == 'percentage'):
                            if (isinstance(item, (float, int))):
                                item = StatsProcessor.cal_non_percentage(item, to_str=True, postfix="%")
                        elif ("YoY" in item_name):
                            if (isinstance(item, (float, int))):
                                item = StatsProcessor.cal_percentage(item)
                        else:
                            if (isinstance(item, (float, int))):
                                item = StatsProcessor.cal_non_percentage(item,  postfix="千元")
                        table_dict[index_name][(time_index, item_name)] = item

                    except KeyError:
                        if (index_name not in table_dict.keys()):
                            table_dict[index_name] = dict()

                        table_dict[index_name][(time_index, item_name)] = item

        total_table = pd.DataFrame.from_dict(table_dict, orient='index')
        total_table.columns = pd.MultiIndex.from_tuples(total_table.columns)

        for name, setting in self.table_settings.items():
            if ('target_index' in setting.keys()):
                target_indexes = [target_index.strip() for target_index in setting['target_index']]
            else:
                target_indexes = [None]
            for target_index in target_indexes:
                try:
                    return_dict[name] = StatsProcessor.slice_multi_col_table(
                        total_table=total_table,
                        mode=setting['mode'],
                        target_index=target_index)
                    break
                except Exception as e:
                    continue
        return return_dict



import os
from pymongo import MongoClient
from dotenv import load_dotenv

class DBClient:
    def __init__(self, mongo_uri):
        """初始化時接收 MongoDB 連接 URI"""
        self.client = MongoClient(mongo_uri)
        self.db = self.client["twse_company"]
    
    def get_client(self):
        return self.client

    def get_collection(self, collection_name):
        """返回指定名稱的集合"""
        return self.db[collection_name]


from .base import StatsFetcher
from datetime import datetime, timedelta, date
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor


class ValueFetcher(StatsFetcher):

    def __init__(self, ticker: str, db_client):
        super().__init__(ticker, db_client)

    def prepare_query(self, start_date, end_date):
        pipeline = super().prepare_query()

        pipeline.append(
            {
                "$project":
                    {
                        "_id": 0,
                        "ticker": 1,
                        "company_name": 1,
                        "daily_data":
                            {
                                "$map":
                                    {
                                        "input":
                                            {
                                                "$filter":
                                                    {
                                                        "input": "$daily_data",
                                                        "as": "daily",
                                                        "cond":
                                                            {
                                                                "$and":
                                                                    [
                                                                        {
                                                                            "$gte": ["$$daily.date", start_date]
                                                                        }, {
                                                                            "$lte": ["$$daily.date", end_date]
                                                                        }
                                                                    ]
                                                            }
                                                    }
                                            },
                                        "as": "daily_item",
                                        "in":
                                            {
                                                "date": "$$daily_item.date",
                                                "close": "$$daily_item.close",
                                                "P_B": "$$daily_item.P_B",
                                                "P_E": "$$daily_item.P_E",
                                                "P_FCF": "$$daily_item.P_FCF",
                                                "P_S": "$$daily_item.P_S",
                                                "EV_OPI": "$$daily_item.EV_OPI",
                                                "EV_EBIT": "$$daily_item.EV_EBIT",
                                                "EV_EBITDA": "$$daily_item.EV_EBITDA",
                                                "EV_S": "$$daily_item.EV_S"
                                            }
                                    }
                            },
                        "yearly_data": 1
                    }
            })

        return pipeline

    def collect_data(self, start_date, end_date):
        pipeline = self.prepare_query(start_date, end_date)

        fetched_data = list(self.collection.aggregate(pipeline))

        return fetched_data[0]

    def query_data(self):
        try:
            latest_time = StatsDateTime.get_latest_time(self.ticker, self.collection)['last_update_time']
            target_year = latest_time['daily_data']['last_update'].year
            start_date = latest_time['daily_data']['last_update'] - timedelta(days=31)
            end_date = latest_time['daily_data']['last_update']

        except Exception as e:
            today = StatsDateTime.get_today()
            target_year = today.year
            start_date = (today.date - timedelta(days=31))
            end_date = today.date

        this_year = target_year - 1911

        fetched_data = self.collect_data(start_date, end_date)

        fetched_data['daily_data'] = fetched_data['daily_data'][-1]

        fetched_data['yearly_data'] = ValueProcessor.transform_to_df(
            fetched_data['daily_data'],
            fetched_data['yearly_data'],
        )

        return fetched_data

    def query_value_serie(self):
        """
        回傳指定公司的歷來評價
        return : Dict[pd.DataFrame]
        Dict中包含以下八個key, 每個key對應DataFrame
            {
                P_E,
                P_FCF,
                P_S,
                P_B,
                EV_OPI,
                EV_EBIT,
                EV_EBITDA,
                EV_S
            }
        """

        pipeline = [
            {
                "$match": {
                    "ticker": self.ticker,
                }
            },
            {
                "$project":
                    {
                        "_id": 0,
                        "ticker": 1,
                        "company_name": 1,
                        "daily_data":
                            {
                                "$map":
                                    {
                                        "input": "$daily_data",    # 正確地指定要處理的陣列
                                        "as": "daily",    # 每個元素的名稱
                                        "in":
                                            {
                                                "date": "$$daily.date",
                                                "P_E": "$$daily.P_E",
                                                "P_FCF": "$$daily.P_FCF",
                                                "P_B": "$$daily.P_B",
                                                "P_S": "$$daily.P_S",
                                                "EV_OPI": "$$daily.EV_OPI",
                                                "EV_EBIT": "$$daily.EV_EBIT",
                                                "EV_EBITDA": "$$daily.EV_EBITDA",
                                                "EV_S": "$$daily.EV_S"
                                            }
                                    }
                            }
                    }
            }
        ]

        fetched_data = self.collection.aggregate(pipeline).to_list()
        fetched_data = fetched_data[0]

        value_keys = ["P_E", "P_FCF", "P_B", "P_S", "EV_OPI", "EV_EBIT", "EV_EVITDA", "EV_S"]
        return_dict = {value_key: dict() for value_key in value_keys}

        for value_key in value_keys:
            for data in fetched_data['daily_data']:
                if (value_key not in data.keys()):
                    continue
                else:
                    return_dict[value_key].update({data['date']: data[value_key]})

        return_dict = {
            value_key: pd.DataFrame.from_dict(value_dict, orient='index', columns=[value_key])
            for value_key, value_dict in return_dict.items()
        }
        return return_dict


class ValueProcessor(StatsProcessor):

    @staticmethod
    def transform_to_df(daily_dict, yearly_dict):
        latest_data = {"year": f"過去4季"}

        latest_data.update(daily_dict)
        latest_data.pop("date")
        latest_data.pop("close")

        yearly_dict.append(latest_data)

        yearly_dict = pd.DataFrame.from_dict(yearly_dict)

        return yearly_dict



from .base import StatsFetcher
from datetime import datetime, timedelta
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import importlib.resources as pkg_resources
import yaml


class MarginTradingFetcher(StatsFetcher):

    def __init__(self, ticker, db_client):
        """
        iFa -> 交易資訊 -> 資券變化

        包括: 
        1. 當日交易
        2. 一年內交易
        """
        super().__init__(ticker, db_client)

    def prepare_query(self, start_date, end_date):
        pipeline = super().prepare_query()

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "daily_data": {
                    "$map": {
                        "input": {
                            "$filter": {
                                "input": "$daily_data",
                                "as": "daliy",
                                "cond": {
                                    "$and": [{
                                        "$gte": ["$$daliy.date", start_date]
                                    }, {
                                        "$lte": ["$$daliy.date", end_date]
                                    }]
                                }
                            }
                        },
                        "as": "target_daliy_data",
                        "in": "$$target_daliy_data"
                    }
                },
                "margin_trading": {
                    "$map": {
                        "input": {
                            "$filter": {
                                "input": "$margin_trading",
                                "as": "margin",
                                "cond": {
                                    "$and": [{
                                        "$gte": ["$$margin.date", start_date]
                                    }, {
                                        "$lte": ["$$margin.date", end_date]
                                    }]
                                }
                            }
                        },
                        "as": "target_margin_data",
                        "in": "$$target_margin_data"
                    }
                },
                "security_lending": {
                    "$map": {
                        "input": {
                            "$filter": {
                                "input": "$security_lending",
                                "as": "lending",
                                "cond": {
                                    "$and": [{
                                        "$gte": ["$$lending.date", start_date]
                                    }, {
                                        "$lte": ["$$lending.date", end_date]
                                    }]
                                }
                            }
                        },
                        "as": "target_lending_data",
                        "in": "$$target_lending_data"
                    }
                }
            }
        })

        return pipeline

    def collect_data(self, start_date, end_date):
        pipeline = self.prepare_query(start_date, end_date)

        fetched_data = self.collection.aggregate(pipeline).to_list()

        return fetched_data[-1]

    def query_data(self):
        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection)['last_update_time']
            latest_date = latest_time['margin_trading']['latest_date']
            end_date = latest_date.replace(hour=0,
                                           minute=0,
                                           second=0,
                                           microsecond=0)
        except Exception as e:
            print(
                f"No updated time for institution_trading in {self.ticker}, use current time instead"
            )
            end_date = datetime.now(self.timezone)
            end_date = end_date.replace(hour=0,
                                        minute=0,
                                        second=0,
                                        microsecond=0)

            if (end_date.hour < 22):  # 拿不到今天的資料
                end_date = end_date - timedelta(days=1)

        start_date = end_date - timedelta(days=365)

        fetched_data = self.collect_data(start_date, end_date)

        fetched_data['daily_data'] = sorted(fetched_data['daily_data'],
                                            key=lambda x: x['date'],
                                            reverse=True)
        fetched_data['margin_trading'] = sorted(fetched_data['margin_trading'],
                                                key=lambda x: x['date'],
                                                reverse=True)
        fetched_data['security_lending'] = sorted(
            fetched_data['security_lending'],
            key=lambda x: x['date'],
            reverse=True)

        table_dict = self.process_data(fetched_data)

        return table_dict

    def process_data(self, fetched_data):
        return_dict = dict()

        daily_datas = fetched_data['daily_data']
        latest_data = fetched_data['daily_data'][0]
        yesterday_data = fetched_data['daily_data'][1]

        # 交易價格與昨天交易
        price_dict = {
            "open": latest_data['open'],
            'close': latest_data['close'],
            'range': f"{latest_data['low']} - {latest_data['high']}",
            'volume': latest_data['volume'] / 1000,
            'last_open': yesterday_data['open'],
            'last_close': yesterday_data['close'],
            'last_range':
            f"{yesterday_data['low']} - {yesterday_data['high']}",
            'last_volume': yesterday_data['volume'] / 1000
        }
        annual_lows = [data['low'] for data in daily_datas]
        annual_highs = [data['high'] for data in daily_datas]
        lowest = np.min(annual_lows).item()
        highest = np.max(annual_highs).item()
        price_dict['52weeks_range'] = f"{lowest} - {highest}"

        return_dict['price'] = price_dict

        # 當日交易
        margin_trading = fetched_data['margin_trading']
        security_lending = fetched_data['security_lending']

        latest_margin_date = margin_trading[0]['date']
        latest_lending_date = security_lending[0]['date']
        ## 融資融券
        ### 先將所有現金償還與現券償還改成現償
        for trading in margin_trading:
            trading['financing']['現償'] = trading['financing'].pop('現金償還')
            trading['short_selling']['現償'] = trading['short_selling'].pop(
                '現券償還')
        ### 轉換
        latest_margin_trading = margin_trading[0]
        latest_margin_trading_df = {
            category: sub_dict
            for category, sub_dict in latest_margin_trading.items()
            if (isinstance(sub_dict, dict))
        }
        latest_margin_trading_df = pd.DataFrame.from_dict(
            latest_margin_trading_df)

        ## 借券表格
        latest_stock_lending = security_lending[0]['stock_lending']

        latest_stock_lending = {
            type_name: StatsProcessor.cal_round_int(value / 1000)
            for type_name, value in latest_stock_lending.items()
        }
        latest_stock_lending.pop("前日餘額")
        latest_stock_lending_df = pd.DataFrame.from_dict(latest_stock_lending,
                                                         orient="index",
                                                         columns=['stock_lending'])

        latest_dict = {
            "date": latest_margin_date,
            "margin_trading": latest_margin_trading_df,
            "stock_lending": latest_stock_lending_df,
            "security_offset": latest_margin_trading['資券互抵']
        }

        return_dict['latest_trading'] = latest_dict

        # 一年內
        annual_dates = [
            data['date'].strftime('%Y-%m-%d')
            for data in fetched_data['margin_trading']
        ]
        close_prices = {
            data['date'].strftime('%Y-%m-%d'): data['close']
            for data in fetched_data['daily_data']
            if data['date'].strftime('%Y-%m-%d') in annual_dates
        }
        volumes = {
            data['date'].strftime('%Y-%m-%d'):
            StatsProcessor.cal_round_int(data['volume'] / 1000)
            for data in fetched_data['daily_data']
            if data['date'].strftime('%Y-%m-%d') in annual_dates
        }

        ## 融資融券
        financings = {
            data['date'].strftime("%Y-%m-%d"): {
                f"融資_{type_name}": num_of_stocks
                for type_name, num_of_stocks in data['financing'].items()
            }
            for data in fetched_data['margin_trading']
        }

        short_sellings = {
            data['date'].strftime("%Y-%m-%d"): {
                f"融券_{type_name}": num_of_stocks
                for type_name, num_of_stocks in data['short_selling'].items()
            }
            for data in fetched_data['margin_trading']
        }

        ### 資券互抵
        security_offsets = {
            data['date'].strftime("%Y-%m-%d"): data['資券互抵']
            for data in fetched_data['margin_trading']
        }

        ## 借券(stock lendings)
        stock_lendings = {
            data['date'].strftime('%Y-%m-%d'): {
                f"借券_{type_name}":
                (num_of_stocks /
                 1000) if isinstance(num_of_stocks,
                                     (int, float)) else num_of_stocks
                for type_name, num_of_stocks in data['stock_lending'].items()
            }
            for data in fetched_data['security_lending']
        }

        annual_dict = {
            date: {
                "close": close_prices[date],
                "volume": volumes[date],
                **financings[date],
                **short_sellings[date],
                **stock_lendings[date], 
                "資券互抵":security_offsets[date]
            }
            for date in annual_dates
        }

        annual_table = pd.DataFrame.from_dict(annual_dict)

        return_dict['annual_margin'] = annual_table.T

        return return_dict



from .base import StatsFetcher
import pandas as pd
import yfinance as yf


class TechFetcher(StatsFetcher):

    def __init__(self, ticker: str, db_client):
        """
        The Capitalization-Weighted Index includes the following tickers:
        ['GSPC', 'IXIC', 'DJI', 'TWII']   
        """

        super().__init__(ticker, db_client)
        self.collection = self.db["TWN/APIPRCD"]
        self.full_ohlcv = self._get_ohlcv()
        self.basic_indexes = [
            'SMA5', 'SMA20', 'SMA60', 'EMA5', 'EMA20', 'EMA40', 'EMA12',
            'EMA26', 'RSI7', 'RSI14', 'RSI21', 'MACD', 'Signal Line',
            'Middle Band', 'Upper Band', 'Lower Band', '%b', 'ATR', 'BBW',
            'EMA Cycle', 'EMA Cycle Instructions', 'Day Trading Signal'
        ]

        self.daily_index = TechProcessor.cal_basic_index(self.full_ohlcv)

        self.weekly_index = TechProcessor.resample(
            self.daily_index,
            period='W',
            technical_indicators=self.basic_indexes
        )

        self.monthly_index = TechProcessor.resample(
            self.daily_index,
            period='ME',
            technical_indicators=self.basic_indexes
        )

        self.quarterly_index = TechProcessor.resample(
            self.daily_index,
            period='QE',
            technical_indicators=self.basic_indexes
        )

        self.yearly_index = TechProcessor.resample(
            self.daily_index,
            period='YE',
            technical_indicators=self.basic_indexes
        )

    def _get_ohlcv(self):

        required_cols = ['date', 'open', 'high', 'low', 'close', 'volume']

        try:
            # 先對yf search
            if self.ticker in ['GSPC', 'IXIC', 'DJI', 'TWII']:
                full_tick = f'^{self.ticker}'
            else:
                full_tick = f'{self.ticker}.tw'

            df = self.conduct_yf_search(full_tick)

            if not self.has_required_columns(df, required_cols):

                print(f".tw failed, try .two")

                full_tick = f'{self.ticker}.two'

                df = self.conduct_yf_search(full_tick)

                if (df.empty):
                    raise ValueError(f"No data found for ticker: {self.ticker}")

        except (KeyError, ValueError, TypeError) as e:
            # 再對TEJ search
            tej_required_cols = [
                "mdate", "open_d", 'high_d', 'low_d', 'close_d', 'vol'
            ]
            tej_name_proj = {
                tej_name: org_name
                for tej_name, org_name in zip(tej_required_cols, required_cols)
            }

            query = {'ticker': self.ticker}
            ticker_full = self.collection.find_one(query)

            if not ticker_full:
                raise ValueError("No ticker found in database")

            daily_data = ticker_full.get("data", [])
            if not isinstance(daily_data, list):
                raise TypeError("Expected 'daily_data' to be a list.")

            df = pd.DataFrame(daily_data)

            if not self.has_required_columns(df, tej_required_cols):
                raise KeyError(f"Missing required columns")
            df = df.rename(columns=tej_name_proj)

        return df[required_cols]

    def get_daily(self):

        return self.daily_index

    def get_weekly(self):

        return self.weekly_index

    def get_monthly(self):

        return self.monthly_index

    def get_quarterly(self):

        return self.quarterly_index

    def get_yearly(self):

        return self.yearly_index

    def conduct_yf_search(self, ticker: str):

        yf_ticker = yf.Ticker(ticker)
        origin_df = yf_ticker.history(period="10y")

        if origin_df.empty:
            return origin_df

        origin_df = origin_df.reset_index()
        origin_df["Date"] = pd.to_datetime(origin_df["Date"])
        df = origin_df.rename(
            columns={
                "Date": "date",
                "Open": "open",
                "High": "high",
                "Low": "low",
                "Close": "close",
                "Volume": "volume"
            }
        )

        return df


class TechProcessor:

    @staticmethod
    def cal_sma(closes: pd.Series, n_days: int) -> pd.Series:
        return closes.rolling(window=n_days).mean()

    @staticmethod
    def cal_ema(closes: pd.Series, n_days: int) -> pd.Series:
        return closes.ewm(span=n_days, adjust=False).mean()

    @staticmethod
    def cal_rsi(closes: pd.Series, n_days: int) -> pd.Series:
        delta = closes.diff(1)
        gain = (delta.where(delta > 0, 0)).rolling(window=n_days).mean()
        loss = (-delta.where(delta < 0, 0)).rolling(window=n_days).mean()
        rs = gain / loss
        return 100 - (100 / (1 + rs))

    @staticmethod
    def cal_macd(ema12s: pd.Series, ema26s: pd.Series) -> pd.Series:
        return ema12s - ema26s

    @staticmethod
    def cal_single_line(macds: pd.Series, n_days: int = 9) -> pd.Series:
        return macds.ewm(span=n_days, adjust=False).mean()

    @staticmethod
    def cal_bollinger_bands(
        closes: pd.Series, n_days: int = 20
    ) -> pd.DataFrame:
        middle = closes.rolling(window=n_days).mean()
        upper = middle + 2 * closes.rolling(window=n_days).std()
        lower = middle - 2 * closes.rolling(window=n_days).std()
        percent_b = (closes - lower) / (upper - lower)

        return pd.DataFrame(
            {
                'middle': middle,
                'upper': upper,
                "lower": lower,
                "%b": percent_b
            }
        )

    @staticmethod
    def cal_atr(
        highes: pd.Series, lows: pd.Series, closes: pd.Series, n_days: int
    ) -> pd.Series:
        high_low = highes - lows
        high_close = (highes - closes.shift(1)).abs()
        low_close = (lows - closes.shift(1)).abs()

        true_range = pd.concat([high_low, high_close, low_close],
                               axis=1).max(axis=1)
        atr = true_range.rolling(window=n_days, min_periods=1).mean()

        return atr

    @staticmethod
    def check_tech_trend(ema5: float, ema20: float, ema40: float) -> str:
        if ema5 > ema20 > ema40:
            return '穩定上升期'
        elif ema20 > ema5 > ema40:
            return '牛市結束期'
        elif ema20 > ema40 > ema5:
            return '熊市入口期'
        elif ema40 > ema20 > ema5:
            return '穩定下跌期'
        elif ema40 > ema5 > ema20:
            return '熊市結束期'
        elif ema5 > ema40 > ema20:
            return '牛市入口期'
        else:
            return '未定義'

    @staticmethod
    def check_day_trading(
        close_today: float,
        close_yesterday: float,
        today_atr: float,
        today_rsi7: float,
        target_value=270,
        atr_thred=7.5
    ) -> str:
        reasons = []

        # 檢查規則 1: Close 今日 - Close 昨日
        if close_today - close_yesterday <= 0:
            reasons.append('今日收盤價未高於昨日收盤價')

        # 檢查規則 2: 今日 Close 是否大於目標值
        if close_today < target_value:
            reasons.append(f'今日收盤價未達到目標值{target_value}')

        # 檢查規則 3: ATR 是否大於 atr_thred
        if today_atr < atr_thred:
            reasons.append(f'ATR 值小於 {atr_thred}')

        # 檢查規則 4: RSI7 是否大於等於 40
        if today_rsi7 < 40:
            reasons.append('RSI7 值小於 40')

        # 根據檢查結果返回
        if not reasons:
            return '今日此股票為好的當沖標的'
        else:
            return f'今日此股票並非好的當沖標的, 原因: {", ".join(reasons)}'

    @staticmethod
    def cal_basic_index(ohlcvs: pd.DataFrame):

        # SMA
        ohlcvs['SMA5'] = TechProcessor.cal_sma(ohlcvs['close'], 5)
        ohlcvs['SMA20'] = TechProcessor.cal_sma(ohlcvs['close'], 20)
        ohlcvs['SMA60'] = TechProcessor.cal_sma(ohlcvs['close'], 40)

        # EMA
        ohlcvs['EMA5'] = TechProcessor.cal_ema(ohlcvs['close'], 5)
        ohlcvs['EMA20'] = TechProcessor.cal_ema(ohlcvs['close'], 20)
        ohlcvs['EMA40'] = TechProcessor.cal_ema(ohlcvs['close'], 40)

        ohlcvs['EMA12'] = TechProcessor.cal_ema(ohlcvs['close'], 12)
        ohlcvs['EMA26'] = TechProcessor.cal_ema(ohlcvs['close'], 26)

        # RSI
        ohlcvs['RSI7'] = TechProcessor.cal_rsi(ohlcvs['close'], 7)
        ohlcvs['RSI14'] = TechProcessor.cal_rsi(ohlcvs['close'], 14)
        ohlcvs['RSI21'] = TechProcessor.cal_rsi(ohlcvs['close'], 21)

        # MACD
        ohlcvs['MACD'] = TechProcessor.cal_macd(
            ohlcvs['EMA12'], ohlcvs['EMA26']
        )
        ohlcvs['Signal Line'] = TechProcessor.cal_single_line(ohlcvs['MACD'], 9)

        # BANDS
        bands = TechProcessor.cal_bollinger_bands(ohlcvs['close'], 20)
        ohlcvs['Middle Band'] = bands['middle']
        ohlcvs['Upper Band'] = bands['upper']
        ohlcvs['Lower Band'] = bands['lower']
        ohlcvs['%b'] = bands['%b']
        ohlcvs['BBW'] = (ohlcvs["Upper Band"] -
                         ohlcvs["Lower Band"]) / ohlcvs["Middle Band"]

        # ATR
        ohlcvs['ATR'] = TechProcessor.cal_atr(
            ohlcvs['high'], ohlcvs['low'], ohlcvs['close'], 14
        )

        # EMA CYCLE
        ohlcvs['EMA Cycle'] = ohlcvs.apply(
            lambda row: TechProcessor.
            check_tech_trend(row['EMA5'], row['EMA20'], row['EMA40']),
            axis=1
        )
        guidance_map = {
            '穩定上升期': "三條移動平均線都左下右上, 買方優勢, 三線間隔越來越遠時, 進一步強攻",
            '牛市結束期':
            "ema20 & 40 左下右上, ema5 緩慢下滑, 行情仍強, 賣出條件為 ema5 持續下跌, ema20 停止上漲",
            '熊市入口期': "全數出清穩定上升期布局的多頭部位, 考慮提早佈局建立空頭部位",
            '穩定下跌期': "三條移動平均線都是左上右下, 賣方優勢, 三線間隔越來越遠時, 進一步強攻",
            '熊市結束期':
            "ema20 & 40 左上右下, ema5 緩慢上升, 行情仍走弱, 布局買進的條件是 ema 持續上漲, ema20 停止下降, 幾乎持平",
            '牛市入口期': "全數出清穩定下跌期布局的空頭部位, 考慮提早佈局多頭部位",
            '未定義': "無對應指導"
        }

        ohlcvs['EMA Cycle Instructions'] = ohlcvs['EMA Cycle'].map(guidance_map)

        # DAY TRADE SELECTING
        ohlcvs['close_yesterday'] = ohlcvs['close'].shift(1)
        ohlcvs['Day Trading Signal'] = ohlcvs.apply(
            lambda row: TechProcessor.check_day_trading(
                close_today=row['close'],
                close_yesterday=row['close_yesterday'],    # 使用前一天的收盤價
                today_atr=row['ATR'],
                today_rsi7=row['RSI7']
            ),
            axis=1
        )

        return ohlcvs

    @staticmethod
    def resample(
        df: pd.DataFrame,
        period='W',
        technical_indicators=None,
        date_col='date'
    ):
        """
        將 DataFrame 中的技術指標數據重新取樣為指定的時間週期。
        參數:
        df (pd.DataFrame): 包含技術指標的日線 DataFrame，必須包含 datetime 索引或指定的日期列。
        period (str): 重新取樣的時間週期，如 'W', 'ME', 'QE', 'YE'。
        technical_indicators (list): 需要重新取樣的技術指標列名。
        date_col (str): 包含日期的列名（如果不是索引）。
        
        返回:
        pd.DataFrame: 重新取樣後的 DataFrame。
        """
        # 標記輸入的日期是否是索引還是列
        original_index = df.index.name == date_col if df.index.name else False

        # 檢查是否需要將日期列設置為索引
        if not original_index and date_col in df.columns:
            df[date_col] = pd.to_datetime(df[date_col])
            df.set_index(date_col, inplace=True)

        # 過濾掉非數字型的列
        numeric_df = df.select_dtypes(include='number')

        # 預設的聚合方式
        agg_dict = {
            'open': 'first',
            'high': 'max',
            'low': 'min',
            'close': 'last',
            'volume': 'sum'
        }

        # 如果沒有提供 technical_indicators，設置默認技術指標
        if technical_indicators is None:
            technical_indicators = numeric_df.columns.tolist()

        # 將技術指標的聚合方式設置為 'mean'
        for indicator in technical_indicators:
            if indicator in numeric_df.columns:
                agg_dict[indicator] = 'mean'

        # 過濾出存在於 DataFrame 中的列
        existing_cols = {
            col: agg_dict[col]
            for col in agg_dict if col in numeric_df.columns
        }

        # 確保索引為 DatetimeIndex，進行重新取樣
        if not isinstance(df.index, pd.DatetimeIndex):
            raise TypeError(
                "The DataFrame index must be a DatetimeIndex for resampling."
            )

        resampled_df = numeric_df.resample(period).agg(existing_cols)

        # 如果原始日期是列而非索引，將日期從索引還原為列並重置索引為範圍索引
        if not original_index:
            resampled_df.reset_index(inplace=True)

        return resampled_df



from .base import StatsFetcher, StatsDateTime
import importlib.resources as pkg_resources
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import yaml



class ProfitLoseFetcher(StatsFetcher):
    """
    iFa.ai: 財務分析 -> 損益表
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.table_settings = StatsProcessor.load_yaml("profit_lose.yaml")

    def prepare_query(self, target_season):
        pipeline = super().prepare_query()

        pipeline.append({
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "profit_loses": {
                    "$sortArray": {
                        "input": {
                            "$map": {
                                "input": {
                                    "$filter": {
                                        "input": "$seasonal_data",
                                        "as": "season",
                                        "cond": {
                                            "$eq":
                                            ["$$season.season", target_season]
                                        }
                                    }
                                },
                                "as": "target_season_data",
                                "in": {
                                    "year":
                                    "$$target_season_data.year",
                                    "season":
                                    "$$target_season_data.season",
                                    "profit_lose":
                                    "$$target_season_data.profit_lose"
                                }
                            }
                        },
                        "sortBy": {
                            "year": -1
                        }  # 按 year 降序排序
                    }
                }
            }
        })

        return pipeline

    def collect_data(self, target_season):
        pipeline = self.prepare_query(target_season)

        fetched_data = self.collection.aggregate(pipeline)

        return list(fetched_data)[-1]

    def query_data(self):
        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection)['last_update_time']
            target_season = latest_time['seasonal_data']['latest_season']
        except Exception as e:
            today = StatsDateTime.get_today()

            target_season = today.season
            target_season = target_season - 1 if target_season > 1 else 4

        fetched_data = self.collect_data(target_season)

        return self.process_data(fetched_data, target_season)

    def process_data(self, fetched_data, target_season):

        profit_loses = fetched_data['profit_loses']

        index_names = []

        table_dict = dict()
        grand_total_dict = dict()

        return_dict = {
            "ticker": fetched_data['ticker'],
            "company_name": fetched_data['company_name'],
        }

        for data in profit_loses:
            year = data['year']

            time_index = f"{year}Q{target_season}"

            # 蒐集整體的keys
            index_names += list(data['profit_lose'].keys())
            profit_lose = data['profit_lose']

            for index_name, value_dict in profit_lose.items():
                # (2020Q1, 項目, 金額或%)
                for item_name, item in value_dict.items():
                    if ('percentage' in item_name):
                        if (isinstance(item, (float, int))):
                            item = StatsProcessor.cal_non_percentage(item, to_str=True, postfix="%")
                    elif ('YoY' in item_name):
                        if (isinstance(item, (float, int))):
                            item = StatsProcessor.cal_percentage(item)
                    elif ('每股盈餘' in index_name):
                        if (isinstance(item, (float, int))):
                            item = StatsProcessor.cal_non_percentage(item,  postfix="元")
                    else:
                        if (isinstance(item, (float, int))):
                            item = StatsProcessor.cal_non_percentage(item,  postfix="千元")
                    try:
                        table_dict[index_name][(time_index, item_name)] = item

                    except KeyError:
                        if (index_name not in table_dict.keys()):
                            table_dict[index_name] = dict()
                            grand_total_dict[index_name] = dict()

                        table_dict[index_name][(time_index, item_name)] = item

        total_table = pd.DataFrame.from_dict(table_dict, orient='index')
        total_table.columns = pd.MultiIndex.from_tuples(total_table.columns)

        total_table = total_table.replace("N/A", None)

        for name, setting in self.table_settings.items():
            if ('target_index' in setting.keys()):
                target_indexes = [target.strip() for target in setting['target_index']]
            else:
                target_indexes = [None]
            
            for target_index in target_indexes:
                try:
                    return_dict[name] = StatsProcessor.slice_multi_col_table(
                        total_table=total_table,
                        mode=setting['mode'],
                        target_index=target_index)
                    break
                except Exception as e:
                    return_dict[name] = StatsProcessor.slice_multi_col_table(
                        total_table=total_table,
                        mode=setting['mode'],
                        target_index=target_index)

        return return_dict



from setuptools import setup, find_packages

setup(
    name='neurostats_API',
    version='0.0.21b',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "numpy",
        "pandas",
        "pymongo",
        "pytz",
        "python-dotenv",
        "yfinance"
    ],
    author='JasonWang@Neurowatt',
    packages=find_packages(exclude=['.venv', 'test*', 'data_in_db', '.pytest_cache']),
    include_package_data=True,
    # package_dir={'neurostats_service': 'src'},
    package_data={'neurostats_API': ['tools/*.yaml', 'tools/*/*.yaml', 'tools/*.txt']},
    author_email='jason@neurowatt.ai',
    description='The service of NeuroStats website',
    url='https://github.com/NeurowattStats/NeuroStats_API.git',
    python_requires='>=3.6'
)
